################################################################################
#   Gene prediction pipeline 
#
#   $Id: PredictionParser.py 2764 2009-09-04 16:55:07Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
PredictionParser.py - Parser for exonerate/genewise output
==========================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

"""
import os, sys, string, re, getopt, tempfile, copy

import Genomics, GTF
try: import alignlib
except ImportError: pass

## number of nucleotides to ignore for counting stop-codons
## This used to be three, but is now set to 0 in order to keep
## accounting for stop codons consistent with the disruptions
## table.
GlobalBorderStopCodons = 0
GlobalStopCodons = ("TAG", "TAA", "TGA")

class Error(Exception):
    """Base class for exceptions in this module."""
    def __str__(self):
        return str(self.message)
    def _get_message(self, message): return self._message
    def _set_message(self, message): self._message = message
    message = property(_get_message, _set_message)

class InputError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class AlignmentError(Error):
    """Exception raised by inconsistent alignments.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class PredictionParserEntry :

    def __init__(self, expand = 1):

        self.mExpand = expand
        
        self.mPredictionId = 0
        self.mQueryToken = 0
        self.mQueryFrom = 0
        self.mQueryTo = 0
        self.mSbjctToken = 0
        self.mSbjctStrand = 0
        self.mSbjctFrom = 0
        self.mSbjctTo = 0
        self.mRank = 0
        self.score = 0
        self.mQueryLength = 0
        self.mQueryCoverage = 0
        self.mNGaps = 0
        self.mNFrameShifts = 0
        self.mNIntrons = 0
        self.mNSplits = 0
        self.mNStopCodons = 0
        self.mPercentIdentity = 0
        self.mPercentSimilarity = 0
        self.mTranslation = ""
        self.mSbjctGenomeFrom = 0
        self.mSbjctGenomeTo = 0
        self.mAlignmentString = ""
        self.mQueryAli = ""
        self.mSbjctAli = ""
        
        if self.mExpand:
            self.mMapPeptide2Translation = alignlib.makeAlignmentVector()
            self.mMapPeptide2Genome = []
        else:
            self.mMapPeptide2Translation = None
            self.mMapPeptide2Genome = None
        self.mNAssembled = 0

    def __del__(self):
        if self.mExpand:
            self.mMapPeptide2Translation.clear()
            self.mMapPeptide2Genome = []
        self.mTranslation = ""
        self.mAlignmentString = ""
        
    def __eq__(self, other ):

        return self.mQueryToken == other.mQueryToken and \
               self.mSbjctToken == other.mSbjctToken and \
               self.mSbjctStrand == other.mSbjctStrand and \
               self.score == other.score and \
               self.mQueryFrom == other.mQueryFrom and \
               self.mQueryTo == other.mQueryTo and \
               self.mSbjctGenomeFrom == other.mSbjctGenomeFrom and \
               self.mSbjctGenomeTo == other.mSbjctGenomeTo and \
               self.mAlignmentString == other.mAlignmentString 

    def CleanAlignmentInfo( self ):
        """remove alignment information (to save space)."""
        self.mTranslation, self.mAlignmentString, self.mQueryAli, self.mSbjctAli = None, None, None, None
        
    def GetCopy( self ):
        """return a new copy.
        """

        new_entry = PredictionParserEntry()

        new_entry.mExpand = self.mExpand 
        
        new_entry.mPredictionId = self.mPredictionId 
        new_entry.mQueryToken = self.mQueryToken 
        new_entry.mQueryFrom = self.mQueryFrom 
        new_entry.mQueryTo = self.mQueryTo 
        new_entry.mSbjctToken = self.mSbjctToken 
        new_entry.mSbjctStrand = self.mSbjctStrand 
        new_entry.mSbjctFrom = self.mSbjctFrom 
        new_entry.mSbjctTo = self.mSbjctTo 
        new_entry.mRank = self.mRank 
        new_entry.score = self.score 
        new_entry.mQueryLength = self.mQueryLength 
        new_entry.mQueryCoverage = self.mQueryCoverage 
        new_entry.mNGaps = self.mNGaps 
        new_entry.mNFrameShifts = self.mNFrameShifts 
        new_entry.mNIntrons = self.mNIntrons 
        new_entry.mNSplits = self.mNSplits 
        new_entry.mNStopCodons = self.mNStopCodons 
        new_entry.mPercentIdentity = self.mPercentIdentity 
        new_entry.mPercentSimilarity = self.mPercentSimilarity 
        new_entry.mTranslation = self.mTranslation 
        new_entry.mSbjctGenomeFrom = self.mSbjctGenomeFrom 
        new_entry.mSbjctGenomeTo = self.mSbjctGenomeTo 
        new_entry.mAlignmentString = self.mAlignmentString 
        new_entry.mQueryAli = self.mQueryAli 
        new_entry.mSbjctAli = self.mSbjctAli 

        if self.mExpand:
            new_entry.mMapPeptide2Translation = alignlib.makeAlignmentVector()
            alignlib.copyAlignment( new_entry.mMapPeptide2Translation, self.mMapPeptide2Translation)
            new_entry.mMapPeptide2Genome = Genomics.String2Alignment( new_entry.mAlignmentString) 
        else:
            new_entry.mMapPeptide2Translation = self.mMapPeptide2Translation = None
            new_entry.mMapPeptide2Genome = self.mMapPeptide2Genome = None

        return new_entry

    def Add( self, const_other,
             combine_contig = False,
             allow_overlap = False,
             contig_size = 0):
        """add one entry to another.

        This procedure allows to add
        - predictions on different contigs if combine_contig = True
        - overlapping predictions on the same query if allow_overlap = True
        """

        other = const_other.GetCopy()

        if not self.mExpand or not other.mExpand:
            raise ValueError, "both object must be expanded."
        
        if self.mQueryToken != other.mQueryToken:
            raise ValueError, "can not add different query."

        if self.mSbjctToken != other.mSbjctToken or \
               self.mSbjctStrand != other.mSbjctStrand :
            if combine_contig:
                self.mSbjctToken += "-" + other.mSbjctToken
                self.mSbjctStrand += other.mSbjctStrand
            else:
                raise ValueError, "can not add different sbjct."                

        query_overlap = max( 0, min(self.mQueryTo, other.mQueryTo) -\
                             max(self.mQueryFrom, other.mQueryFrom) + 1)
        
        if query_overlap > 0:
            
            if allow_overlap:
                overlap = query_overlap
                ## if queries overlap, truncate this before adding the other
                self.mMapPeptide2Translation.removeRowRegion( self.mQueryTo - overlap + 1, self.mQueryTo )
                other.mMapPeptide2Translation.moveAlignment( 0, -overlap )
                self.mQueryTo -= overlap
                self.mTranslation = self.mTranslation[:-overlap]
                
                ## remove aligned residues from the back
                for x in range(len(self.mMapPeptide2Genome) - 1, 0, -1):
                    if self.mMapPeptide2Genome[x][1] <= overlap:
                        overlap -= self.mMapPeptide2Genome[x][1]
                        del self.mMapPeptide2Genome[x]
                    else:
                        break
                self.mMapPeptide2Genome[-1] = (self.mMapPeptide2Genome[-1][0],
                                               self.mMapPeptide2Genome[-1][1] - overlap,
                                               self.mMapPeptide2Genome[-1][2] - overlap * 3)
            else:
                raise ValueError, "can not add overlapping entries: overlap = %i, queries:\n%s\n%s\n" % (query_overlap, str(self), str(other))

        sbjct_overlap = max(0, min(self.mSbjctGenomeTo, other.mSbjctGenomeTo) -\
                            max(self.mSbjctGenomeFrom, other.mSbjctGenomeFrom), 0)

        if sbjct_overlap > 0:
            if not combine_contig:
                raise ValueError, "can not add overlapping entries: overlap = %i, sbjct:\n%s\n%s\n" % (sbjct_overlap, str(self), str(other))
        
        if self.mSbjctToken == other.mSbjctToken:
            ## join on the same contig
            if self.mSbjctGenomeFrom < other.mSbjctGenomeFrom:
                ## get length of gap
                d_na = other.mSbjctGenomeFrom - self.mSbjctGenomeTo
                d_aa = other.mQueryFrom       - self.mQueryTo - 1
                self.mMapPeptide2Genome += [("P", d_aa, d_na)] + other.mMapPeptide2Genome
                self.mTranslation += other.mTranslation
                other.mMapPeptide2Translation.moveAlignment( 0, self.mSbjctTo - 1 )
            else:
                d_na = self.mSbjctGenomeFrom - other.mSbjctGenomeTo
                d_aa = self.mQueryFrom - other.mQueryTo - 1
                self.mTranslation = other.mTranslation + self.mTranslation
                self.mMapPeptide2Genome = other.mMapPeptide2Genome + [("P", d_aa, d_na)] + self.mMapPeptide2Genome
                self.mMapPeptide2Translation.moveAlignment( 0, other.mSbjctTo - 1 )
                
            self.mSbjctGenomeFrom = min(self.mSbjctGenomeFrom, other.mSbjctGenomeFrom )
            self.mSbjctGenomeTo = max(self.mSbjctGenomeTo, other.mSbjctGenomeTo )        
                
        else:
            ## join on different contigs
            d_na = contig_size - self.mSbjctGenomeTo + other.mSbjctGenomeFrom + query_overlap * 3
            d_aa = other.mQueryFrom - self.mQueryTo - 1
            self.mMapPeptide2Genome += [("P", d_aa, d_na),] + other.mMapPeptide2Genome
            self.mTranslation += other.mTranslation 
            other.mMapPeptide2Translation.moveAlignment( 0, self.mSbjctTo - 1 )

            self.mSbjctGenomeFrom = self.mSbjctGenomeFrom
            self.mSbjctGenomeTo = contig_size + other.mSbjctGenomeTo

        self.mRank = max( self.mRank, other.mRank)
        self.score += other.score
        self.mQueryCoverage += other.mQueryCoverage
        self.mNGaps += other.mNGaps
        self.mNFrameShifts += other.mNFrameShifts
        self.mNIntrons += other.mNIntrons + 1
        self.mNStopCodons += other.mNStopCodons

        nself  = self.mMapPeptide2Translation.getLength() - self.mMapPeptide2Translation.getNumGaps()
        nother = other.mMapPeptide2Translation.getLength() - other.mMapPeptide2Translation.getNumGaps()
        alignlib.addAlignment2Alignment( self.mMapPeptide2Translation, other.mMapPeptide2Translation )

        self.mQueryFrom = self.mMapPeptide2Translation.getRowFrom()
        self.mQueryTo = self.mMapPeptide2Translation.getRowTo()
        self.mSbjctFrom = self.mMapPeptide2Translation.getColFrom()
        self.mSbjctTo = self.mMapPeptide2Translation.getColTo()
        
        nnew = self.mMapPeptide2Translation.getLength() - self.mMapPeptide2Translation.getNumGaps()
        
        self.mPercentIdentity = (self.mPercentIdentity * nself + other.mPercentIdentity * nother) / nnew
        self.mPercentSimilarity = (self.mPercentSimilarity * nself + other.mPercentSimilarity * nother) / nnew

        self.mAlignmentString = string.join( map( \
                                      lambda x: string.join(map(str, x), " "),
                                      self.mMapPeptide2Genome), " ")

        self.mNAssembled += 1 + other.mNAssembled

    def InvertGenomicCoordinates( self, lgenome ):
        """invert genomic alignment on sequence.

        Negative strand is calculated from the other end.
        """
        if self.mSbjctStrand == "-":
            x = self.mSbjctGenomeFrom 
            self.mSbjctGenomeFrom = lgenome - self.mSbjctGenomeTo
            self.mSbjctGenomeTo   = lgenome - x
        
    def ShiftGenomicRegion( self, offset ):
        """shift aligned genomic region by offset."""
        self.mSbjctGenomeFrom += offset
        self.mSbjctGenomeTo += offset

    def ShiftQueryRegion( self, offset ):
        """shift aligned genomic region by offset."""
        self.mQueryFrom += offset
        self.mQueryTo += offset
        if self.mExpand:
            self.mMapPeptide2Translation.moveAlignment( offset, 0)

    def SetStrand( self, strand ):
        """set strand."""
        self.mSbjctStrand = strand 

    def __str__( self ):
        """get a string representation of results."""
            
        if self.mExpand:
            if self.mMapPeptide2Translation.getLength() > 0:
                f = alignlib.AlignmentFormatEmissions( self.mMapPeptide2Translation )
                row_ali, col_ali = f.mRowAlignment, f.mColAlignment
                self.mQueryFrom = self.mMapPeptide2Translation.getRowFrom()
                self.mQueryTo = self.mMapPeptide2Translation.getRowTo()
                self.mSbjctFrom = self.mMapPeptide2Translation.getColFrom()
                self.mSbjctTo = self.mMapPeptide2Translation.getColTo()
            else:
                row_ali, col_ali = "", ""
        else:
            row_ali = self.mQueryAli
            col_ali = self.mSbjctAli
            
        if self.mPredictionId:
            return string.join( map(str, (\
                self.mPredictionId,
                self.mQueryToken,
                self.mSbjctToken,
                self.mSbjctStrand,
                self.mRank,
                self.score,
                self.mQueryFrom,
                self.mQueryTo,
                row_ali,
                self.mSbjctFrom,
                self.mSbjctTo,
                col_ali,
                self.mQueryLength,
                self.mQueryCoverage, 
                self.mNGaps,
                self.mNFrameShifts,
                self.mNIntrons,
                self.mNSplits,
                self.mNStopCodons,
                self.mPercentIdentity,
                self.mPercentSimilarity,
                self.mTranslation,
                self.mSbjctGenomeFrom,
                self.mSbjctGenomeTo,
                self.mAlignmentString,
                self.mNAssembled,
                )), "\t")
        else:
            return string.join( map(str, (\
              self.mQueryToken,
              self.mSbjctToken,
              self.mSbjctStrand,
              self.mRank,
              self.score,
              self.mQueryFrom,
              self.mQueryTo,
              row_ali,
              self.mSbjctFrom,
              self.mSbjctTo,
              col_ali,
              self.mQueryLength,
              self.mQueryCoverage, 
              self.mNGaps,
              self.mNFrameShifts,
              self.mNIntrons,
              self.mNSplits,
              self.mNStopCodons,
              self.mPercentIdentity,
              self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom,
              self.mSbjctGenomeTo,
              self.mAlignmentString,
              self.mNAssembled,
              )), "\t")

    def FillFromTable( self, table_row ):

        if len(table_row) == 25:
            ( self.mPredictionId,
              self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              self.mAlignmentString) = table_row
        elif len(table_row) == 26:
            ( self.mPredictionId,
              self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              self.mAlignmentString,
              self.mNAssembled) = table_row[:26]
        elif len(table_row) > 26:
            ( self.mPredictionId,
              self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              self.mAlignmentString,
              self.mNAssembled) = table_row[:26]
        else:
            raise ValueError, "unknown format: %i fields" % len(data)
            sys.exit(0)
            
        if self.mExpand:
            self.mMapPeptide2Translation = alignlib.makeAlignmentVector()

            if self.mQueryAli != "" and self.mSbjctAli != "":
                f = alignlib.AlignmentFormatEmissions()
                f.mRowFrom = self.mQueryFrom
                f.mColFrom = self.SbjctFrom
                f.mRowAlignment = self.mQueryAli
                f.mColAlignment = self.mSbjctAli

                f.copy( self.mMapPeptide2Translation )

            self.mMapPeptide2Genome = Genomics.String2Alignment( self.mAlignmentString )

    def SetTranslation( self, genomic_sequence ):
        """set translation from genomic sequence."""
        self.mMapPeptide2Translation, self.mTranslation = Genomics.Alignment2PeptideAlignment( \
                self.mMapPeptide2Genome, self.mQueryFrom, self.mSbjctGenomeFrom, genomic_sequence )
        
    def GetExonBoundariesOnGenome( self ):
        """return a list of exons for this prediction.
        """
        return map( lambda x: (x[3], x[4]),
                    Genomics.Alignment2ExonBoundaries( self.mMapPeptide2Genome,
                                                       self.mQueryFrom,
                                                       self.mSbjctGenomeFrom ))
        
    def Write( self ):
        """write a tab separated list of results.
        """
        print str(self)
        
    def Read( self, line ):

        data = string.split( line[:-1], "\t")
        
        if len(data) == 26:
            ( self.mPredictionId,
              self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              self.mAlignmentString, self.mNAssembled,
              ) = data
        elif len(data) == 25:
            ( self.mPredictionId,
              self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              self.mAlignmentString,
              ) = data
        elif len(data) == 24:            
            ( self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              self.mAlignmentString,
              ) = data
        elif len(data) == 23:
            ( self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
              self.mRank, self.score,
              self.mQueryFrom, self.mQueryTo, self.mQueryAli,
              self.mSbjctFrom, self.mSbjctTo, self.mSbjctAli,
              self.mQueryLength, self.mQueryCoverage, 
              self.mNGaps, self.mNFrameShifts, self.mNIntrons,
              self.mNSplits, self.mNStopCodons,
              self.mPercentIdentity, self.mPercentSimilarity,
              self.mTranslation,
              self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
              ) = data
            self.mAlignmentString = ""
        else:
            raise ValueError, "unknown format: %i fields in line %s" % (len(data), line[:-1])

        (self.score, self.mQueryCoverage, self.mPercentIdentity, self.mPercentSimilarity) = map (\
            float, (self.score, self.mQueryCoverage, self.mPercentIdentity, self.mPercentSimilarity))

        (self.mQueryFrom, self.mQueryTo, self.mQueryLength,
         self.mSbjctFrom, self.mSbjctTo,
         self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
         self.mNGaps, self.mNIntrons, self.mNSplits, self.mNStopCodons,
         self.mNFrameShifts, self.mNAssembled) = map (\
            int, ( self.mQueryFrom, self.mQueryTo, self.mQueryLength,
                   self.mSbjctFrom, self.mSbjctTo,
                   self.mSbjctGenomeFrom, self.mSbjctGenomeTo,
                   self.mNGaps, self.mNIntrons, self.mNSplits, self.mNStopCodons,
                   self.mNFrameShifts, self.mNAssembled))

        if self.mExpand:        
            self.mMapPeptide2Translation = alignlib.makeAlignmentVector()

            if self.mQueryAli != "" and self.mSbjctAli != "":
                f = alignlib.AlignmentFormatEmissions()
                f.mRowFrom = self.mQueryFrom
                f.mColFrom = self.mSbjctFrom
                f.mRowAlignment = self.mQueryAli
                f.mColAlignment = self.mSbjctAli

                f.copy( self.mMapPeptide2Translation )

            self.mMapPeptide2Genome = Genomics.String2Alignment( self.mAlignmentString )



##-----------------------------------------------------------------------------------------------
class PredictionParserEntryCaleb( PredictionParserEntry ):
    """a parser to convert Calebs predictions.

    The format is:
    ENSG00000000005 ENST00000003616 688.97  96.85   100     1       318     317     17 61 108 142 193 249   NULL    NULL    951     100     X       +       77495591        77511098      MAKNPPENCEDCHILNAEAFKSKKICKSLKICGLVFGILALTLIVLFWGGKHFWPETPKKTYDMEHTFYSNGEKKKIYMEIDPVTRTEIFRSGNGTDETLEVHDFKNGYTGIYFVGLQKCFIKTQIKVIPEFSEPEEEIDENEEITTTFFEQSVIWVPAEKPIENRDFLKNSKILEICDNVTMYWINPTLIAVSELQDFEEDGEDLHFPTNEKKGIEQNEQWVVPQVKMEKIRHTRQASEEELPINDYTENGLEFDPMLDERGYCCIYCRRGNRYCRRVCEPLLGYYPYPYCYQGGRVICRVIMPCNWWVARMLGRV atggcaaagaatcctccagagaactgtgaggactgtcacattctaaatgcagaagcttttaaatccaagaagatatgtaaatcacttaagatttgtggattggtgtttggtatcctggctctaactctaattgtcctgttttgggggggcaagcacttctggccggagacacccaaaaaaacatatgacatggagcacactttctacagcaatggagagaagaagaagatttacatggaaattgatcctgtgaccagaactgaaatattcagaagtggaaatggcactgatgaaacattggaagtacatgactttaaaaatggttacactggcatctactttgtaggtcttcaaaaatgcttcatcaaaactcagattaaagtgattcctgaattttctgaaccagaggaggaaatagatgagaatgaagaaatcaccacaactttttttgaacagtctgtgatttgggtcccagcagaaaagcctattgaaaatcgagactttctgaaaaattccaaaattctggaaatttgtgataatgtgaccatgtattggatcaatcccactctaatagcagtttcagagttacaagactttgaggaagatggtgaagatcttcactttcctaccaatgaaaaaaaaggtattgaacaaaacgagcagtgggtggtcccccaagtgaagatggaaaagatccgccacaccagacaagcaagtgaggaagaacttccaataaatgactatactgaaaatggactagaatttgatcccatgctggatgagagaggttattgttgtatttactgccgtcgaggcaaccgctactgccgccgcgtctgtgaacctttactaggttactacccgtatccatactgctatcaaggagggcgggttatctgtcgtgtcatcatgccttgcaactggtgggtagcccgcatgctggggagggtc     MAKNPPENCEDCHILNAEAFKSKKICKSLKICGLVFGILALTLIVLFWGSKHFWPEVPKKAYDMEHTFYSNGEKKKIYMEIDPVTRTEIFRSGNGTDETLEVHDFKNGYTGIYFVGLQKCFIKTQIKVIPEFSEPEEEIDENEEITTTFFEQSVIWVPAEKPIENRDFLKNSKILEICDNVTMYWINPTLISVSELQDFEEEGEDLHFPANEKKGIEQNEQWVVPQVKVEKTRHARQASEEELPINDYTENGIEFDPMLDERGYCCIYCRRGNRYCRRVCEPLLGYYPYPYCYQGGRVICRVIMPCNWWVARMLGRV       MAKNPPENCEDCHILNAEAFKSKKICKSLKICGLVFGILALTLIVLFWGGKHFWPETPKKTYDMEHTFYSNGEKKKIYMEIDPVTRTEIFRSGNGTDETLEVHDFKNGYTGIYFVGLQKCFIKTQIKVIPEFSEPEEEIDENEEITTTFFEQSVIWVPAEKPIENRDFLKNSKILEICDNVTMYWINPTLIAVSELQDFEEDGEDLHFPTNEKKGIEQNEQWVVPQVKMEKIRHTRQASEEELPINDYTENGLEFDPMLDERGYCCIYCRRGNRYCRRVCEPLLGYYPYPYCYQGGRVICRVIMPCNWWVARMLGRV       0       3904_results    0       0.0167  0.2019  0.0825  678.5   272.5   0.00517 0.03308 4.8054  -1484.5 0.2092  NULL    0       317     317     1       0    317      317     317     NULL    NULL    794.08

    Note: Caleb's protein positions sometimes start 0.
    
    """
    def __init__(self):
        
        PredictionParserEntry.__init__(self)
        
    def Read( self, line ):

        data = string.split( line[:-1], "\t")

        (self.mQueryGene, self.mQueryToken, self.score,
         self.mPercentIdentity, self.mQueryCoverage,
         self.mQueryFrom, self.mQueryTo, self.mQueryLength,
         introns, self.mNFrameShifts, self.mNStopCodons,
         psl_score, psl_identity, self.mSbjctToken, self.mSbjctStrand, self.mSbjctGenomeFrom,
         self.mSbjctGenomeTo, self.mTranslation, cdna, query_aligned_string, sbjct_aligned_string, errors,
         f_name, self.mRank) = data[:24]
        
        self.score = float(self.score)
        self.mPercentIdentity = float(self.mPercentIdentity)
        self.mQueryCoverage = float(self.mQueryCoverage)

        self.mSbjctToken = "chr%s" % self.mSbjctToken
        
        (self.mQueryFrom, self.mQueryTo, self.mSbjctGenomeFrom, self.mSbjctGenomeTo) = map(int, \
            (self.mQueryFrom, self.mQueryTo, self.mSbjctGenomeFrom, self.mSbjctGenomeTo))

        if self.mNFrameShifts == "NULL":
            self.mNFrameShifts = 0
        else:
            self.mNFrameShifts = len(string.split(self.mNFrameShifts, " "))

        if self.mNStopCodons == "NULL":
            self.mNStopCodons = 0
        else:
            self.mNStopCodons = len(string.split(self.mNStopCodons, " "))
            
        if introns == "NULL":
            self.mNIntrons == 0
        else:
            self.mNIntrons = len(string.split(introns, " "))

        ## correct bug in parsing: entries like A:F[ttt] or -:F[ttt] in strings
        self.mTranslation = re.sub( "\s*\S{0,1}:(\S)\[\S{3}\]\s*", lambda x: x.groups()[0], self.mTranslation)

##--------------------------------------------------------------------------
class Predictions :
    mLogLevel = 0

    def __init__( self ):
        self.mMatches = []

    def __eq__(self, other):
        if not other: return 0        

        if not other.mMatches: return 0
        if not self.mMatches: return 0
        return self.mMatches[0] == other.mMatches[0]

    def __getitem__(self, key ):
        return self.mMatches[key]

    def __setitem__(self, key, value ):
        self.mMatches[key] = value

    def __len__(self):
        return len(self.mMatches)

    def __str__(self):
        return string.join( map(str, self.mMatches), "\n")

    def append( self, entry ):
        return self.mMatches.append( entry )
    
    def Clear( self ):
        for x in self.mMatches:
            x.__del__()
        self.mMatches = []
    
    def Combine(self, other):
        self.mMatches += other.mMatches

    def Add( self, prediction ):
        """prediction to predictions."""
        self.mMatches.append(prediction)

    def ReRank( self, offset = 1 ):
        """rerank predictions based on current sorting order."""
        
        for x in range(len(self.mMatches)):
            self.mMatches[x].mRank = offset + x

    def GetBestMatch( self ):
        return self.mMatches[0]
        
    def GetNumMatches( self ):
        """return number of matches."""
        return len(self.mMatches)

    def ShiftGenomicRegion( self, offset ):
        """shift aligned genomic region by offset for all entries."""
        for x in self.mMatches:
            x.ShiftGenomicRegion( offset )

    def ShiftQueryRegion( self, offset ):
        """shift aligned genomic region by offset for all entries."""
        for x in self.mMatches:
            x.ShiftQueryRegion( offset )

    def SetStrand( self, strand ):
        """set strand."""
        for x in self.mMatches:
            x.SetStrand( strand )
    
    def Write( self ):
        """write a tab separated list of results.
        """
        for x in self.mMatches:
            x.Write()

    def Sort( self, cmp_function = None ):
        """sort predictions.
        """
        self.mMatches.sort( cmp_function )

    def GetRange( self ):
        """return a prediction covering all predictions."""

        p = self.mMatches[0].GetCopy()

        p.mExpand = False
        
        if len(self.mMatches) == 1: return p
        
        for pp in self.mMatches[1:]:

            p.mQueryFrom       = min(p.mQueryFrom      , pp.mQueryFrom )
            p.mQueryTo         = max(p.mQueryTo        , pp.mQueryTo )
            p.mSbjctGenomeFrom = min(p.mSbjctGenomeFrom, pp.mSbjctGenomeFrom )
            p.mSbjctGenomeTo   = max(p.mSbjctGenomeTo  , pp.mSbjctGenomeTo )
            p.score += pp.score

        return p
##--------------------------------------------------------------------------
class PredictionParser :
    
    mLogLevel = 0
    
    def __init__( self ):
        self.mMatches = []
        self.mErrors = []
        self.mNumInput = 0
        self.mNumOutput = 0
        
    def __eq__(self, other):
        if not other: return 0        

        if not other.mMatches: return 0
        if not self.mMatches: return 0
        return self.mMatches[0] == other.mMatches[0]

    def __getitem__(self, key ):
        return self.mMatches[key]

    def __setitem__(self, key, value ):
        self.mMatches[key] = value

    def __len__(self):
        return len(self.mMatches)

    def __str__(self):
        return string.join( map(str, self.mMatches), "\n")

    def GetNumErrors( self ):
        """return number of errors."""
        return len(self.mErrors)

    def GetNumInput( self ):
        """return number of input."""
        return self.mNumInput

    def GetNumOutput( self ):
        """return number of output."""
        return self.mNumOutput

    def Parse( self, lines, peptide_sequence = None, genomic_sequence = None ):
        """parse lines.
        """
        if self.mLogLevel >= 3:
            print "# received"
            print string.join(lines)

        return self.ExtractData( lines, peptide_sequence = peptide_sequence, genomic_sequence = genomic_sequence)
    
    def Write( self ):
        """write a tab separated list of results.
        """
        for x in self.mMatches:
            x.Write()

    def Sort( self, cmp_function = None ):
        """sort predictions.
        """
        self.mMatches.sort( cmp_function )

def iterator_predictions( infile ):
    """a simple iterator over all entries in a file."""
    while 1:
        line = infile.readline()
        if line.startswith("id\t"): continue
        if not line: return
        if line[0] == "#": continue
        e = PredictionParserEntry()
        e.Read( line )
        yield e

##--------------------------------------------------------------------------
class PredictionParserGenewise( PredictionParser ):
    """Parse output from genewise.
    """
    
    def __init__(self):
        PredictionParser.__init__( self )

        self.mParsers = [ self.ParseSummary, self.ParseTranslation, self.ParsePeptide, self.ParseGFF, self.ParseAlb ]

    def GetNew( self ):
        """return object of the same type."""
        return PredictionParserGenewise()

    def ExtractData( self, lines, peptide_sequence, genomic_sequence ):
        """call the various parsers for a chunk of data.
        """

        if len(lines) == 0:
            return 0

        matches = Predictions()
        self.mMatches = matches
        
        # get info about chromosomal location of chunk submitted to
        # genewise.
        if lines[0][:2] == "#>":
            (self.mQueryToken, self.mSbjctToken, self.mSbjctStrand,
             self.mSbjctFrom, self.mSbjctTo) = re.split( "\s+", lines[0][2:-1])[:5]
            self.mSbjctFrom, self.mSbjctTo = map(int, (self.mSbjctFrom, self.mSbjctTo))
            first_index = 1
        else:
            self.mQueryToken, self.mSbjctToken, self.mSbjctStrand, self.mSbjctFrom, self.mSbjctTo = "", "", "", 0, 0
            first_index = 0

        if len(lines) == 1:
            print "## ERROR: no data for %s vs %s" % (self.mQueryToken, self.mSbjctToken)
            return

        
        entry = PredictionParserEntry()
        entry.mRank = 1
    
        parser_counter = 0

        ## call the various parsers
        for index in range(first_index, len(lines)):

            if lines[index][:2] == "//":
                self.mParsers[parser_counter](lines[first_index:index])

                parser_counter += 1
                first_index = index + 1

        if first_index < index:
            entry.mParsers[parser_counter](lines[first_index:index]) 

        ## set the fields according to parsed output
        # fill in summary info for each entry:
        for entry in self.mMatches:
            entry.mQueryToken = self.mQueryToken
            entry.mSbjctToken = self.mSbjctToken            
            entry.mSbjctStrand = self.mSbjctStrand

            if peptide_sequence:
                row_seq = alignlib.makeSequence( peptide_sequence )
                col_seq = alignlib.makeSequence( entry.mTranslation )
                alignlib.rescoreAlignment( entry.mMapPeptide2Translation, row_seq, col_seq )
        
                entry.mQueryLength = len(peptide_sequence)            
                entry.mPercentIdentity = alignlib.calculatePercentIdentity( entry.mMapPeptide2Translation, row_seq, col_seq ) * 100
                entry.mPercentSimilarity = alignlib.calculatePercentSimilarity( entry.mMapPeptide2Translation ) * 100
                entry.mQueryCoverage = ( entry.mMapPeptide2Translation.getRowTo() - \
                                         entry.mMapPeptide2Translation.getRowFrom() + 1 ) * 100 /\
                                         entry.mQueryLength

            entry.mAlignmentString = string.join( map( \
                                      lambda x: string.join(map(str, x), " "),
                                      entry.mMapPeptide2Genome), " ")

            if len(entry.mMapPeptide2Genome) == 0:
                print "### PARSING ERROR: empty alignment"
                print str(entry)
                print string.join(lines, "")
                sys.exit(1)
            
            # get regions matching
            # use summary information, as it is correct for genomic data.
            # when -u is set (in contrast to -alb format)
            # genewise starts counting at 1, thus subtract 1 from first position
            # Note: When position starts with a gap, genewise says alignments starts
            # from position 1 while the first aligned residue is position 2
            if entry.mMapPeptide2Genome[0][0] == "G": entry.mQueryFrom += entry.mMapPeptide2Genome[0][1]
            if entry.mMapPeptide2Genome[-1][0] == "G": entry.mQueryTo -= entry.mMapPeptide2Genome[0][1]

            # a sanity check
            if entry.mQueryFrom != entry.mMapPeptide2Translation.getRowFrom():
                print "## PARSING ERROR: wrong start at %s vs %s: %i <> %i" %\
                      (entry.mQueryToken, entry.mSbjctToken,
                       entry.mQueryFrom,
                       entry.mMapPeptide2Translation.getRowFrom() )
                print str(entry)
                print string.join(lines,"")
                row_seq = alignlib.makeSequence( peptide_sequence )
                col_seq = alignlib.makeSequence( entry.mTranslation )
                print alignlib.AlignmentFormatExplicit( entry.mMapPeptide2Translation, row_seq, col_seq )
                sys.exit(1)
            if entry.mQueryTo != entry.mMapPeptide2Translation.getRowTo() and \
                   entry.mMapPeptide2Genome[-1][0] != "G":
                print "## PARSING ERROR: wrong end at %s vs %s: %i <> %i" %\
                  (entry.mQueryToken, entry.mSbjctToken,
                   entry.mQueryTo,
                   entry.mMapPeptide2Translation.getRowTo() )
                print str(entry)                
                print string.join(lines,"")
                row_seq = alignlib.makeSequence( peptide_sequence )
                col_seq = alignlib.makeSequence( entry.mTranslation )
                print alignlib.AlignmentFormatExplicit( entry.mMapPeptide2Translation, row_seq, col_seq )
                sys.exit(1)

            (entry.mNIntrons, entry.mNFrameShifts, entry.mNGaps, entry.mNSplits, entry.mNStopCodons, disruptions) = \
                              Genomics.CountGeneFeatures( entry.mSbjctGenomeFrom,
                                                          entry.mMapPeptide2Genome,
                                                          genomic_sequence,
                                                          GlobalBorderStopCodons,
                                                          GlobalStopCodons )

        matches.Sort( lambda x,y : cmp( -x.score, -y.score))
        return matches
    
    def ParseGFF( self, lines ):
        return "gff", None

    def ParseSummary( self, lines ):
        """parse summary entry of genewise."""

        for line in lines[1:]:
            data = re.split( "\s+", line[:-1])
            entry = PredictionParserEntry()
            entry.score = float(data[0])
            entry.mQueryFrom = int(data[2])
            entry.mQueryTo = int(data[3])
            entry.mSbjctGenomeFrom = int(data[5]) - 1
            entry.mSbjctGenomeTo = int(data[6])
            entry.mIndels = int(data[7])
            if data[8] == "":
                entry.mIntrons = 0
            else:
                entry.mIntrons = int(data[8])
            self.mMatches.append( entry )

    def ParseTranslation( self, lines ):
        """read a translation entry from genewise.

        return a tuple (description, sequence)
        """

        if len(lines) < 1:
            raise "not enough lines"
        
        n = 0
        sequence = ""
        for line in lines:
            if line[0] in (">", "#"):
                if sequence:
                    self.mMatches[n].mTranslation = sequence
                    n += 1
                if line[0] != "#":
                    self.mMatches[n].mDescription = line[1:-1]
                else:
                    self.mMatches[n].mDescription = "pseudogene"                    
                sequence = ""
                continue
            elif line[:6] == "Making":
                continue
            sequence += line[:-1]

        self.mMatches[n].mTranslation = sequence
        
    def ParsePeptide( self, lines ):
        """read a peptide entry from genewise.

        return a tuple (description, sequence, modifications)
        """

        self.ParseTranslation(lines)
        
    def ParseAlb( self, lines ):
        """Parse genewise logical block alignment and return aligment.

        :returns: tuple with ``(code, score, ngaps, nframeshifts,
          nintrons, nphase0, nphase1, nphase2,
          alignment of peptide to genome,
          alignment of peptide to translation)``

        """

        map_genewise2code = \
                          { ( "MATCH_STATE", "CODON") : "M",
                            ( "INTRON_STATE", "5SS_PHASE_0"): "5",
                            ( "INTRON_STATE", "5SS_PHASE_1"): "5",                        
                            ( "INTRON_STATE", "5SS_PHASE_2"): "5",
                            ( "INTRON_STATE", "CENTRAL_INTRON"): "I",
                            ( "MATCH_STATE", "3SS_PHASE_0"): "3",
                            ( "MATCH_STATE", "3SS_PHASE_1"): "3",
                            ( "MATCH_STATE", "3SS_PHASE_2"): "3",                        
                            ( "INSERT_STATE", "3SS_PHASE_0") : "3",
                            ( "INSERT_STATE", "3SS_PHASE_1") : "3",
                            ( "INSERT_STATE", "3SS_PHASE_2") : "3",                                                
                            ( "DELETE_STATE", "INSERT") : "G",
                            ( "INSERT_STATE", "CODON") : "G",
                            ( "INSERT_STATE", "SEQUENCE_DELETION") : "F", # a codon with less than 3 nucleotides at an insert
                            ( "MATCH_STATE", "SEQUENCE_INSERTION") : "F", # a codon with more than 3 nucleotides
                            ( "MATCH_STATE", "SEQUENCE_DELETION")  : "F", # a codon with less than 3 nucleotides
                            ( "LOOP_STATE", "RANDOM_SEQUENCE") : None,
                            ( "END", "END") : None,
                            }
        
        raise "might be broken with new alignlib - check"

        ## note: all the numbers including the ranges can be negative (-1 for first position) !
        rx = re.compile( "^(\S+)\s+\[(\S+):(\S+)\s+\"(\S+)\"\s+(\S+)\],\[(\S+):(\S+)\s+\"(\S+)\"\s+(\S+)\]" )
        
        query_last_state = None
        query_first_from = 0

        sbjct_last_state = None
        sbjct_genomic_residue_first_from = 0
        sbjct_peptide_residue_first_from = 0

        states = []
        total_bin_score = 0.0

        nintrons = 0
        nframeshifts = 0
        nphases = [0,0,0]
        ngaps = 0

        ## parsing errors
        nerrors = 0

        for line in lines:

            try:
                (bin_score,
                 query_from, query_to, query_state, query_score,
                 sbjct_from, sbjct_to, sbjct_state, sbjct_score ) =  rx.match( line[:-1] ).groups()
            except AttributeError:
                raise "parsing error in line %s" % line

            query_from, query_to, sbjct_from, sbjct_to = map(int, (query_from, query_to, sbjct_from, sbjct_to))

            total_bin_score += string.atof( bin_score )

            if query_last_state != query_state or sbjct_last_state != sbjct_state:

                if query_last_state:
                    states.append( ( query_last_state, query_first_from, query_from,
                                     sbjct_last_state, sbjct_first_from, sbjct_from ) )

                
                query_first_from = query_from
                sbjct_first_from = sbjct_from
                query_last_state = query_state
                sbjct_last_state = sbjct_state

        states.append( ( query_last_state, query_first_from, query_to,
                         sbjct_last_state, sbjct_first_from, sbjct_to) )

        ## translate genewise alignment to exonerate alignment codes
        n = 0
        
        ## residue positions on the query/sbjct for peptide sequences
        ## note: genewise would start the first state at -1:0. Thus, add
        ## 1 to get first aligned position in peptide
        query_peptide_from = states[0][1] + 1
        sbjct_peptide_from = 0

        for x in range( len(states)):

            query_state, query_from, query_to, sbjct_state, sbjct_from, sbjct_to = states[x]

            query_len = query_to - query_from
            sbjct_len = sbjct_to - sbjct_from
            
            ## build alignment between peptide sequence and genomic sequence
            try:
                code = map_genewise2code[ (query_state, sbjct_state)]
            except KeyError:
                print "## ERROR: unknown genewise code: %s %s" % (query_state, sbjct_state)
                nerrors += 1
                code = "?"

            if query_state == "LOOP_STATE" and x > 0:
                n += 1
                query_peptide_from = states[x+1][1] + 1
                sbjct_peptide_from = 0
                continue

            ## add code for split codons
            ## move matches to second split codon, if
            ## a split codon is on a gap: add a Gap
            if sbjct_state == "5SS_PHASE_1":
                self.mMatches[n].mMapPeptide2Genome.append( ("S", 0, 1) )
                sbjct_len -= 1
            elif sbjct_state == "5SS_PHASE_2":
                self.mMatches[n].mMapPeptide2Genome.append( ("S", 0, 2) )
                sbjct_len -= 2
            elif sbjct_state == "3SS_PHASE_1":
                query_len = 0
                sbjct_len -= 2
            elif sbjct_state == "3SS_PHASE_2":
                query_len = 0                
                sbjct_len -= 1

            if code:
                if query_from != query_peptide_from - 1:
                    print "Mismatch", query_state, query_from, query_to, sbjct_state,
                    sbjct_from, sbjct_to, query_peptide_from
                
                self.mMatches[n].mMapPeptide2Genome.append( (code, query_len, sbjct_len) )

            if sbjct_state == "3SS_PHASE_1":
                self.mMatches[n].mMapPeptide2Genome.append( ("S", 1, 2) )
                sbjct_len = 2
                if query_state == "MATCH_STATE":
                    query_len = 1
                else:
                    query_len = 0
                    
            elif sbjct_state == "3SS_PHASE_2":
                self.mMatches[n].mMapPeptide2Genome.append( ("S", 1, 1) )                
                sbjct_len = 1
                if query_state == "MATCH_STATE":
                    query_len = 1
                else:
                    query_len = 0
                
            ## build alignment between peptide sequences
            query_increment = 0
            sbjct_increment = 0

            if query_state == "MATCH_STATE":
                 query_increment = query_len

            elif query_state == "DELETE_STATE":
                query_increment = query_len

            elif query_state == "INSERT_STATE":
                query_increment = query_len

            elif query_state == "UNKNOWN_LABEL":
                query_peptide_from += query_len
                ## there are three X's in the sbjct
                sbjct_peptide_from += 3             
                continue
            elif query_state == "LOOP_STATE":
                query_increment += query_len

            if sbjct_state == "CODON":
                sbjct_increment = sbjct_len 
            elif sbjct_state in ("3SS_PHASE_1", "3SS_PHASE_2"):
                sbjct_increment = 3
            elif sbjct_state == "INSERT":
                sbjct_increment = sbjct_len
            elif sbjct_state == "RANDOM_SEQUENCE":
                sbjct_increment = 0

            ## for a frameshift: genewise outputs two characters in the
            ## translated sequence? I do not know. 3 makes more sense.
            elif sbjct_state == "SEQUENCE_DELETION":
                sbjct_increment = 3

            ## for a frameshift: if there is a sequence insertion, no
            ## character is output.
            elif sbjct_state == "SEQUENCE_INSERTION":
                if sbjct_len > 3:
                    sbjct_increment = 3
                else:
                    sbjct_increment = 0

            if code and query_increment and sbjct_increment:
                alignlib.addDiagonal2Alignment(  self.mMatches[n].mMapPeptide2Translation,
                                                 query_peptide_from, query_peptide_from + query_increment,
                                                 sbjct_peptide_from - query_peptide_from
                                                 )
                
            query_peptide_from += query_increment
            sbjct_peptide_from += sbjct_increment / 3

    
            
##--------------------------------------------------------------------------
class PredictionParserExonerate (PredictionParser):
    """Parse output from exonerate.
    """
    
    def __init__(self):
        PredictionParser.__init__( self )

        self.mRestrictToStrand = None

    def SetRestrictToStrand( self, v ):
        self.mRestrictToStrand = v
            
    def ExtractData( self, lines, peptide_sequence = None, genomic_sequence = None):

        matches = Predictions()
        
        nmatches = 0
        for line in lines:
            
            if not re.match("diy", line): continue

            nmatches += 1
            data = string.split( line[:-1], "\t")

            entry = PredictionParserEntry()
            entry.mRank = nmatches 

            (sugar, entry.mQueryLength, rank, percent_identity, percent_similarity, vulgar ) = data[1:]

            (entry.mQueryToken, entry.mQueryFrom, entry.mQueryTo, entry.mQueryStrand, 
             entry.mSbjctToken, entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
             entry.mSbjctStrand, entry.score) = string.split(sugar, " " )

            if self.mRestrictToStrand and entry.mSbjctStrand != self.mRestrictToStrand:
                continue

            try:
                (entry.mQueryFrom, entry.mQueryTo,
                 entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
                 entry.score, entry.mQueryLength) = map( \
                    int , (entry.mQueryFrom, entry.mQueryTo,
                           entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
                           entry.score, entry.mQueryLength))
                
            except ValueError:
                raise InputError, "error while parsing ints: %s" % " ".join( (entry.mQueryFrom, entry.mQueryTo,
                                                                            entry.mSbjctGenomeFrom, entry.mSbjctGenomeTo,
                                                                            entry.score, entry.mQueryLength) )
            
            try:            
                entry.mPercentIdentity, entry.mPercentSimilarity = map( float,
                                                                        (percent_identity, percent_similarity ))
            except ValueError:
                raise InputError, "error while parsing floats: %s, %s" % (percent_identity, percent_similarity)
            
            entry.mAlignmentString = vulgar
            try:
                entry.mMapPeptide2Genome = Genomics.String2Alignment( vulgar )
            except ValueError:
                raise InputError, "error while parsing alignment %s" % vulgar
            
            entry.mQueryCoverage = (entry.mQueryTo - entry.mQueryFrom) * 100 / entry.mQueryLength

            entry.mMapPeptide2Translation, entry.mTranslation = Genomics.Alignment2PeptideAlignment( \
                entry.mMapPeptide2Genome, entry.mQueryFrom, entry.mSbjctGenomeFrom, genomic_sequence )

            ## add 1 for peptide alignments
            entry.mQueryFrom += 1

            (entry.mNIntrons, entry.mNFrameShifts, entry.mNGaps, entry.mNSplits, entry.mNStopCodons, disruptions) = \
                              Genomics.CountGeneFeatures( entry.mSbjctGenomeFrom,
                                                          entry.mMapPeptide2Genome,
                                                          genomic_sequence,
                                                          GlobalBorderStopCodons,
                                                          GlobalStopCodons )

            matches.append(entry)
            
        return matches

##--------------------------------------------------------------------------
def Blocks2Alignment( query_block_starts,
                      sbjct_block_starts,
                      block_sizes ):
    """convert a list of blocks to a peptid2genome alignment.
    """

    alignment = []
    sbjct_pos = 0
    sbjct_offset = sbjct_block_starts[0]
    current_phase = 0

    map_peptide2translation = alignlib.makeAlignmentVector()

    query_peptide_pos = query_block_starts[0]
    sbjct_peptide_pos = 0

    for x in range( len(block_sizes) - 1):

        lquery = block_sizes[x]
        lsbjct = block_sizes[x]
        lgap = query_block_starts[x+1] - ( query_block_starts[x] + lquery )
        lintron = sbjct_block_starts[x+1] - ( sbjct_block_starts[x] + lsbjct * 3 )
        
        ## do peptide2translation alignment
        alignlib.addDiagonal2Alignment( map_peptide2translation,
                                        query_peptide_pos,
                                        query_peptide_pos + lquery,
                                        sbjct_peptide_pos - query_peptide_pos )
        query_peptide_pos += lquery + lgap
        sbjct_peptide_pos += lsbjct 
        
        
        ## do peptide2genome alignment
        lsbjct *= 3
        sbjct_pos += lsbjct
        phase = sbjct_pos % 3
        if phase: lquery -= 1
        
        alignment.append( ("M", lquery, lsbjct) )
        if phase: alignment.append( ("S", 0, phase) )
        alignment.append( ("5", 0, 0) )
        alignment.append( ("I", 0, lintron) )
        alignment.append( ("3", 0, 0) )
        if phase: alignment.append( ("S", 1, 3-phase) )
        if lgap: alignment.append( ("G", lgap, 0) )

    lquery = block_sizes[-1]
    lsbjct = block_sizes[-1]
    alignlib.addDiagonal2Alignment( map_peptide2translation,
                                    query_peptide_pos,
                                    query_peptide_pos + lquery,
                                    sbjct_peptide_pos - query_peptide_pos )


    lsbjct *= 3
    alignment.append( ("M", lquery, lsbjct) )
    return alignment, map_peptide2translation

##--------------------------------------------------------------------------
def Blocks2AlignmentCDNA( query_block_starts,
                          sbjct_block_starts,
                          block_sizes ):
    """convert a list of blocks to a peptid2genome alignment.

    Both blocks are in nucleotides.

    Just get the exon structure correct.
    
    """
        
##     for x in range( len(query_block_starts)):
##         print query_block_starts[x], sbjct_block_starts[x], block_sizes[x]

    alignment = []
    sbjct_pos = 0
    sbjct_offset = sbjct_block_starts[0]
    query_pos = 0
    current_phase = 0

    map_peptide2translation = alignlib.makeAlignmentVector()

    query_peptide_pos = query_block_starts[0]/3
    sbjct_peptide_pos = 0

    query_overhang = 0
    ## phase of last block
    phase = 0
    matches = 0
    splits = 0
    intron = 0
    for x in range( len(block_sizes) - 1):

##         print "x=", x, query_block_starts[x], sbjct_block_starts[x], block_sizes[x],\
##               query_block_starts[x+1] - ( query_block_starts[x] + block_sizes[x] ), \
##               sbjct_block_starts[x+1] - ( sbjct_block_starts[x] + block_sizes[x] )

        # count in nucleotides
        nbases  = block_sizes[x] 
        lgap    = query_block_starts[x+1] - ( query_block_starts[x] + nbases )
        lintron = sbjct_block_starts[x+1] - ( sbjct_block_starts[x] + nbases )

        ## add slipped residues at the ends
        ## add to query, make gap in sbjct smaller.
        ## gaps are thus artificially reduced to multiples of 3
##         nbases  += lgap
##         lsbjct  += lgap
##         lintron -= lgap
##         lgap    -= lgap 

        naminos = (nbases - phase) / 3

        sbjct_pos += nbases
        query_pos += nbases 

        phase = query_pos % 3        

##         print "lgap=", lgap, "lintron=",lintron, "phase=", phase,\
##               "qp=", query_peptide_pos, "sp=", sbjct_peptide_pos, "phase=", phase,\
##               "nbases=",nbases, "naminos=",naminos

        alignment.append( ("M", naminos, naminos * 3) )
        matches += naminos * 3

        # treat gaps in query
        # make sure that gaps only occur at codon positions!
        # if there is overhang (phase > 0) at one more gap.
        # and move position to gap in sbjct
        # new_phase is pseudo-phase, makes sure, that extra
        # nucleotides are taken off from next region.
        new_phase = 0
        if lgap:
            if phase:
                query_pos -= phase
                lgap += phase
                lintron += phase
            # increasing lgap to next codon end
            # remembering to take off residues
            if lgap % 3:
                new_phase = 3 - lgap % 3
                lgap += new_phase
                lintron += new_phase 

            query_pos += lgap
            alignment.append( ("G", lgap / 3, 0) )
            phase = query_pos % 3
            if phase != 0: raise "what?"
            
        if lintron:

            if phase == 0 and lintron % 3 == 0 and lintron < 50:
                alignment.append( ("G", 0, lintron ) )
                intron += lintron
            else:
                if phase: alignment.append( ("S", 0, phase) )
                alignment.append( ("5", 0, 0) )
                alignment.append( ("I", 0, lintron) )
                intron += lintron
                alignment.append( ("3", 0, 0) )

                if phase: alignment.append( ("S", 1, 3-phase) )
                if phase: phase = 3 - phase
                if phase: splits += 1

        if new_phase:
            phase = new_phase
        
        # do peptide2translation alignment
        alignlib.addDiagonal2Alignment( map_peptide2translation,
                                        query_peptide_pos,
                                        query_peptide_pos + naminos,
                                        sbjct_peptide_pos - query_peptide_pos)
        
        query_peptide_pos += naminos + lgap / 3
        sbjct_peptide_pos += naminos

        if phase:
            map_peptide2translation.add( query_peptide_pos,
                                         sbjct_peptide_pos,
                                         0 )
            query_peptide_pos += 1
            sbjct_peptide_pos += 1
        
    nbases = block_sizes[-1]
    
    naminos = (nbases - phase) / 3

    alignlib.addDiagonal2Alignment( map_peptide2translation,
                                    query_peptide_pos,
                                    query_peptide_pos + naminos,
                                    sbjct_peptide_pos - query_peptide_pos )


    sbjct_pos += nbases
    query_pos += nbases 

    naminos = (nbases-phase)/ 3
    phase = query_pos % 3

    alignment.append( ("M", naminos, naminos * 3 ) )
    matches += naminos * 3

##     print "intron=", intron    
##     print "final phase=" , phase
##     print "splits=", splits
##     print "matches=", matches, matches / 3
##     print "query_pos=", query_pos
##     print "sum=", splits * 3 + intron + matches
    return alignment, map_peptide2translation

##--------------------------------------------------------------------------
class PredictionParserBlatCDNA (PredictionParser):
    """Parse output from psl.

    Sample line::

      psLayout version 3

      match   mis-    rep.    N's     Q gap   Q gap   T gap   T gap   strand  Q               Q       Q       Q       T               T       T       T       block   blockSizes
              qStarts  tStarts
              match   match           count   bases   count   bases           name            size    start   end     name            size    start   end     count
      :: ---------------------------------------------------------------------------------------------------------------------------------------------------------------
      885     0       0       0       0       0       3       85      +       ENSCAFT00000000002      885     0       885     chr1.fa 124897793       3228799 3229769 4       41,577,103,164,       0,41,618,721,   3228799,3228845,3229427,3229605,        aactactttggattcacccacagtggggcggcggcagcagc,ggctgcggcccagtatagccagcatccagcttctggtgtagcctactctcatccaactacagttgctagctacgctatccatcaggctccagtagctgctcacacagttactgcggcccatgcaccagcaggcaccacagttgcagttgccaggcctgccccagtagctgttgcagctgctgcaatagctgctgcttatggaggctaccccactggacatacagcaagtgactatggctctatccagagacaacaagaagcaccaccaccagcacccccagctactacacagaactaccaggactcatactcatatgtaagatccactgctcctgctgtagcatatgatagtaagcagctgcctgcccagcctcagccttctgttgctgaaacctactatcagattgcccccaaagcaggttatagccaaggtgcaactcagtatacacaagcccagcaaactcgacaagtgacagccataaaaccagccacaccaagtccagctaccactactttctccatttatcccgtatcttccatcgttcagtcagtagcagttgcagctactgtggtgcca,tatacccagagtgctacttatagtaccatggcagttacttattctggtacatcttattcaggttatgaagcagcagtatattcagctgcatccttctactacc,cagctgcctggacagggaccatctttactaaaaaagcaccattccaaaataaacaactgaaaccaaaacagcctcccaaaccggcccagatagattatcgtgatgtttgtaagattagctgtgctgaaccacagacttataaagaacatttagaaggacaaaaa,     aactactttggattcacccacagtggggcggcggcagcagc,ggctgcggcccagtatagccagcatccagcttctggtgtagcctactctcatccaactacagttgctagctacgctatccatcaggctccagtagctgctcacacagttactgcggcccatgcaccagcaggcaccacagttgcagttgccaggcctgccccagtagctgttgcagctgctgcaatagctgctgcttatggaggctaccccactggacatacagcaagtgactatggctctatccagagacaacaagaagcaccaccaccagcacccccagctactacacagaactaccaggactcatactcatatgtaagatccactgctcctgctgtagcatatgatagtaagcagctgcctgcccagcctcagccttctgttgctgaaacctactatcagattgcccccaaagcaggttatagccaaggtgcaactcagtatacacaagcccagcaaactcgacaagtgacagccataaaaccagccacaccaagtccagctaccactactttctccatttatcccgtatcttccatcgttcagtcagtagcagttgcagctactgtggtgcca,tatacccagagtgctacttatagtaccatggcagttacttattctggtacatcttattcaggttatgaagcagcagtatattcagctgcatccttctactacc,cagctgcctggacagggaccatctttactaaaaaagcaccattccaaaataaacaactgaaaccaaaacagcctcccaaaccggcccagatagattatcgtgatgtttgtaagattagctgtgctgaaccacagacttataaagaacatttagaaggacaaaaa,

    Run is cDNA versus genome.
    """
    
    def __init__(self):
        PredictionParser.__init__( self )
            
    def ExtractData( self, lines, peptide_sequence, genomic_sequence):
        """extract data from psl, tab-formatted output.
        """

        matches = Predictions()
        
        nmatches = 0
        for line in lines:

            if line[0] == "#": continue
            if not re.match("^[0-9]", line): continue

            entry = PredictionParserEntry()
            entry.mRank = nmatches 

            data = string.split(line[:-1], "\t")
            
            ( nmatches, nmismatches, nrepmatches, nns,
              query_ngaps_counts, query_ngaps_bases,
              sbjct_ngaps_counts, sbjct_ngaps_bases,
              sbjct_strand,
              entry.mQueryToken, query_length, query_from, query_to, 
              entry.mSbjctToken, sbjct_length, sbjct_from, sbjct_to, 
              nblocks, block_sizes,
              query_block_starts, sbjct_block_starts,
              query_block_seqs, sbjct_block_seqs) = data

            ( nmatches, nmismatches, nrepmatches, nns,
              query_ngaps_counts, query_ngaps_bases,
              sbjct_ngaps_counts, sbjct_ngaps_bases,
              query_length, query_from, query_to, 
              sbjct_length, sbjct_from, sbjct_to,
              nblocks ) = map(int, \
                              ( nmatches, nmismatches, nrepmatches, nns,
                                query_ngaps_counts, query_ngaps_bases,
                                sbjct_ngaps_counts, sbjct_ngaps_bases,
                                query_length, query_from, query_to, 
                                sbjct_length, sbjct_from, sbjct_to,
                                nblocks ))


            ## compile peptide2genome alignment
            query_block_starts = map(int, string.split( query_block_starts[:-1], ","))
            sbjct_block_starts = map(int, string.split( sbjct_block_starts[:-1], ","))
            query_block_seqs = string.split( query_block_seqs[:-1], ",")
            sbjct_block_seqs = string.split( sbjct_block_seqs[:-1], ",")            
            block_sizes = map(int, string.split( block_sizes[:-1], ","))
            
            if sbjct_strand == "-":
                q = []
                s = []
                for x in range(len(query_block_starts)):
                    q.append( query_length - query_block_starts[x] - block_sizes[x])
                    s.append( sbjct_length - sbjct_block_starts[x] - block_sizes[x])
                q.reverse()
                s.reverse()
                query_block_starts = q
                sbjct_block_starts = s
                block_sizes.reverse()
                sbjct_block_seqs.reverse()
                query_block_seqs.reverse()

            # correct for starts not at codon
            if query_from % 3:
                correct_query = 3 - query_from % 3
                query_block_starts[0] += correct_query
                sbjct_block_starts[0] += correct_query
                block_sizes[0] -= correct_query
                query_from += correct_query
                sbjct_from += correct_query
                print "## correcte from to %i" % query_from
            # correct for stop not at codon
            if query_to % 3:
                correct_query = query_to % 3
                block_sizes[-1] -= correct_query
                query_to -= correct_query
                sbjct_to -= correct_query
                print "## correcte to to %i" % query_to
            
            sbjct_ali_length = sbjct_to - sbjct_from
            query_ali_length = (query_to - query_from) / 3
                            
            query_from /= 3 + 1
            query_to /= 3
            query_length /= 3

            entry.mNIntrons = nblocks - 1
            entry.mQueryLength = query_length 
            entry.mQueryCoverage = 100 * (query_to - query_from) / query_length
            entry.mSbjctStrand = sbjct_strand
            entry.mNGaps = query_ngaps_counts + sbjct_ngaps_counts
            entry.mQueryFrom = query_from
            entry.mQueryTo = query_to
            
            if sbjct_strand == "+":
                entry.mSbjctGenomeFrom = sbjct_from
                entry.mSbjctGenomeTo = sbjct_to
            else:
                entry.mSbjctGenomeFrom = sbjct_length - sbjct_to
                entry.mSbjctGenomeTo = sbjct_length - sbjct_from
            
            entry.mMapPeptide2Genome, entry.mMapPeptide2Translation = \
                                      Blocks2AlignmentCDNA( query_block_starts, sbjct_block_starts, block_sizes )
            entry.mAlignmentString = Genomics.Alignment2String( entry.mMapPeptide2Genome )

            entry.mTranslation = string.join( sbjct_block_seqs, "")
            entry.mSbjctFrom = 1
            entry.mSbjctTo = len(entry.mTranslation)
            
            lquery, lsbjct = Genomics.GetAlignmentLength( entry.mMapPeptide2Genome )
            
            if 0:
                print str(entry)
                if 0 and (lquery != (entry.mQueryTo - entry.mQueryFrom + 1) or lquery != query_ali_length):
                    raise AlignmentError( "alignment length discrepancy in query: %i vs %i vs %i" %\
                                      ( lquery, entry.mQueryTo - entry.mQueryFrom + 1, query_ali_length))

                elif lsbjct != (entry.mSbjctGenomeTo - entry.mSbjctGenomeFrom) or lsbjct != sbjct_ali_length:
                    raise AlignmentError( "# alignment length discrepancy in sbjct: %i vs %i vs %i" %\
                                          ( lsbjct, entry.mSbjctGenomeTo - entry.mSbjctGenomeFrom, sbjct_ali_length))

                elif 0 and (entry.mQueryFrom != entry.mMapPeptide2Translation.getRowFrom() or\
                         entry.mQueryTo != entry.mMapPeptide2Translation.getRowTo()):
                    raise AlignmentError( "# alignment length discrepancy in query: %i-%i vs %i-%i" % \
                                          ( entry.mQueryFrom, entry.mQueryTo,
                                            entry.mMapPeptide2Translation.getRowFrom(),
                                            entry.mMapPeptide2Translation.getRowTo()) )

                elif entry.mSbjctFrom != entry.mMapPeptide2Translation.getColFrom() or\
                         entry.mSbjctTo != entry.mMapPeptide2Translation.getColTo():
                    raise AlignmentError( "# alignment length discrepancy in sbjct: %i-%i vs %i-%i" % \
                                          ( entry.mSbjctFrom, entry.mSbjctTo,
                                            entry.mMapPeptide2Translation.getColFrom(),
                                            entry.mMapPeptide2Translation.getColTo()))

            entry.mPercentIdentity = 100 * nmatches / (nmatches + nmismatches)
            entry.mPercentSimilarity = 100 * nmatches / (nmatches + nmismatches)                

##             peptide_sequence = ["X" * (query_block_starts[0]) + query_block_seqs[0]]
##             p = query_block_starts[0] + block_sizes[0]
            
##             for x in range(1,len(query_block_starts)):
##                 peptide_sequence.append("X" * (query_block_starts[x] - p) + query_block_seqs[x])
##                 p = query_block_starts[x] + block_sizes[x]
                
##             peptide_sequence = string.join( peptide_sequence, "")
##             row_seq = alignlib.makeSequence( peptide_sequence )
##             col_seq = alignlib.makeSequence( entry.mTranslation )
##             alignlib.rescoreAlignment( entry.mMapPeptide2Translation, row_seq, col_seq )
        
##             entry.mPercentIdentity = alignlib.calculatePercentIdentity( entry.mMapPeptide2Translation, row_seq, col_seq ) * 100
##             entry.mPercentSimilarity = alignlib.calculatePercentSimilarity( entry.mMapPeptide2Translation ) * 100
        
            matches.append( entry )

        return matches

    
class PredictionParserBlatTrans (PredictionParser):
    """Parse output from psl.

    Query is protein, database is 6frame translation
    
    Input looks like this::

        psLayout version 3

        match   mis-    rep.    N's     Q gap   Q gap   T gap   T gap   strand  Q               Q       Q       Q       T               T       T       T       block   blockSizes
                qStarts  tStarts
                match   match           count   bases   count   bases           name            size    start   end     name            size    start   end     count
        :: ---------------------------------------------------------------------------------------------------------------------------------------------------------------
        98      0       0       0       0       0       2       3113    +-      185061  98      0       98      chr19.fa        56914383        6068186 6071593 3       28,27,43,        0,28,55,        50842790,50845539,50846068,     IFSCRQNCVEFYPIFLVTLWMAGWYFNQ,VFATCLGLVYIYARHQYFWGYSEAAKK,RITGFRLSLGCLALLTVLGALGIANSFLDEYLDLNVIKKLRHF,   IFSCRQNCVEFYPIFLVTLWMAGWYFNQ,VFATCLGLVYIYARHQYFWGYSEAAKK,RITGFRLSLGCLALLTVLGALGIANSFLDEYLDLNVIKKLRHF,
    """
    
    def __init__(self):
        PredictionParser.__init__( self )
            
    def ExtractData( self, lines, peptide_sequence, genomic_sequence):
        """extract data from psl, tab-formatted output.
        """

        matches = Predictions()
        
        nmatches = 0
        for line in lines:

            if line[0] == "#": continue
            if not re.match("^[0-9]", line): continue
            
            entry = PredictionParserEntry()
            entry.mRank = nmatches 

            data = string.split(line[:-1], "\t")
            
            ( nmatches, nmismatches, nrepmatches, nns,
              query_ngaps_counts, query_ngaps_bases,
              sbjct_ngaps_counts, sbjct_ngaps_bases,
              sbjct_strand,
              entry.mQueryToken, query_length, query_from, query_to, 
              entry.mSbjctToken, sbjct_length, sbjct_from, sbjct_to, 
              nblocks, block_sizes,
              query_block_starts, sbjct_block_starts,
              query_block_seqs, sbjct_block_seqs) = data

            ( nmatches, nmismatches, nrepmatches, nns,
              query_ngaps_counts, query_ngaps_bases,
              sbjct_ngaps_counts, sbjct_ngaps_bases,
              query_length, query_from, query_to, 
              sbjct_length, sbjct_from, sbjct_to,
              nblocks ) = map(int, \
                              ( nmatches, nmismatches, nrepmatches, nns,
                                query_ngaps_counts, query_ngaps_bases,
                                sbjct_ngaps_counts, sbjct_ngaps_bases,
                                query_length, query_from, query_to, 
                                sbjct_length, sbjct_from, sbjct_to,
                                nblocks ))

            entry.score = nmatches
            entry.mNIntrons = nblocks - 1
            entry.mQueryLength = query_length 
            entry.mQueryCoverage = 100 * (query_to - query_from) / query_length
            ## take sbjct strand
            entry.mSbjctStrand = sbjct_strand[1]
            entry.mNGaps = query_ngaps_counts + sbjct_ngaps_counts
            entry.mQueryFrom = query_from + 1
            entry.mQueryTo = query_to

            ## reverse strand according to our coordinates
            if entry.mSbjctStrand == "+":
                entry.mSbjctGenomeFrom = sbjct_from
                entry.mSbjctGenomeTo = sbjct_to
            else:
                entry.mSbjctGenomeFrom = sbjct_length - sbjct_to
                entry.mSbjctGenomeTo = sbjct_length - sbjct_from


            ## compile peptide2genome alignment
            query_block_starts = map(int, string.split( query_block_starts[:-1], ","))
            sbjct_block_starts = map(int, string.split( sbjct_block_starts[:-1], ","))
            query_block_seqs = string.split( query_block_seqs[:-1], ",")
            sbjct_block_seqs = string.split( sbjct_block_seqs[:-1], ",")            
            block_sizes = map(int, string.split( block_sizes[:-1], ","))

            entry.mMapPeptide2Genome, entry.mMapPeptide2Translation = \
                                      Blocks2Alignment( query_block_starts, sbjct_block_starts, block_sizes )
            entry.mAlignmentString = Genomics.Alignment2String( entry.mMapPeptide2Genome )

            entry.mTranslation = string.join( sbjct_block_seqs, "")
            entry.mSbjctFrom = 1
            entry.mSbjctTo = len(entry.mTranslation)

            lquery, lsbjct = Genomics.GetAlignmentLength( entry.mMapPeptide2Genome )
            error = None
            if lquery != (entry.mQueryTo - entry.mQueryFrom + 1):
                raise AlignmentError( "alignment length discrepancy in query: %i vs %i vs %i" %\
                                  ( query_length,  lquery, entry.mQueryTo - entry.mQueryFrom + 1))

            elif lsbjct != (entry.mSbjctGenomeTo - entry.mSbjctGenomeFrom):
                raise AlignmentError( "# alignment length discrepancy in sbjct: %i vs %i vs %i" %\
                                      ( sbjct_length, lsbjct, entry.mSbjctGenomeTo - entry.mSbjctGenomeFrom))
                                                                                                 
                                                                                                 
            elif entry.mQueryFrom != entry.mMapPeptide2Translation.getRowFrom() or\
                     entry.mQueryTo != entry.mMapPeptide2Translation.getRowTo():
                raise AlignmentError( "# alignment length discrepancy in query: %i-%i vs %i-%i" % \
                                      ( entry.mQueryFrom, entry.mQueryTo,
                                        entry.mMapPeptide2Translation.getRowFrom(),
                                        entry.mMapPeptide2Translation.getRowTo()) )
            elif entry.mSbjctFrom != entry.mMapPeptide2Translation.getColFrom() or\
                     entry.mSbjctTo != entry.mMapPeptide2Translation.getColTo():
                raise AlignmentError( "# alignment length discrepancy in sbjct: %i-%i vs %i-%i" % \
                                      ( entry.mSbjctFrom, entry.mSbjctTo,
                                        entry.mMapPeptide2Translation.getColFrom(),
                                        entry.mMapPeptide2Translation.getColTo()))

            peptide_sequence = ["X" * (query_block_starts[0]) + query_block_seqs[0]]
            p = query_block_starts[0] + block_sizes[0]
            
            for x in range(1,len(query_block_starts)):
                peptide_sequence.append("X" * (query_block_starts[x] - p) + query_block_seqs[x])
                p = query_block_starts[x] + block_sizes[x]
                
            peptide_sequence = string.join( peptide_sequence, "")
            row_seq = alignlib.makeSequence( peptide_sequence )
            col_seq = alignlib.makeSequence( entry.mTranslation )
            alignlib.rescoreAlignment( entry.mMapPeptide2Translation, row_seq, col_seq )
        
            entry.mPercentIdentity = alignlib.calculatePercentIdentity( entry.mMapPeptide2Translation, row_seq, col_seq ) * 100
            entry.mPercentSimilarity = alignlib.calculatePercentSimilarity( entry.mMapPeptide2Translation ) * 100
            
            matches.append( entry )

        return matches

class PredictionParserExons (PredictionParser):
    """Parse output from exons file.
    
    Input looks like this::

       CG9926-RA       chr3R   -1      0       1       0       696     9776126 9776819
       CG9927-RA       chr3R   -1      0       1       0       398     9771125 9771523
       CG9927-RA       chr3R   -1      1       2       398     1026    9770442 9771067
       CG9928-RA       chr2L   -1      0       1       0       321     13131417        13131735
       CG9929-RA       chr3R   -1      0       1       0       117     9769026 9769143
       CG9929-RA       chr3R   -1      0       2       117     347     9768742 9768972
       CG9929-RA       chr3R   -1      1       3       347     942     9768091 9768683
       CG9930-RA       chr3R   -1      0       1       0       880     9699040 9699920
       CG9930-RA       chr3R   -1      2       2       880     1038    9694339 9694497
       CG9930-RA       chr3R   -1      0       3       1038    1575    9693241 9693775

    Negative strand exons are given as positive strand coordinates, but are still sorted
    in increasing order.

    Coordinates are zero-based in open/closed notation.

    The score is set to the length of the transcript.
    """
    
    def __init__(self, contig_sizes = {}):
        PredictionParser.__init__( self )
        self.contigSizes = contig_sizes
        
    def ExtractData( self,
                     lines, peptide_sequence,
                     genomic_sequence ):
        """extract data from exons file, tab-formatted output.
        """

        matches = Predictions()
        
        nmatches = 0
        
        entry = None

        id = 0 
        last_query = None
        exons = []
        for line in lines:

            data = line[:-1].split()

            (query, sbjct_token, sbjct_strand,
             phase, exon_id, peptide_from, peptide_to,
             sbjct_genome_from, sbjct_genome_to ) = data

            ( sbjct_genome_from, sbjct_genome_to,
              peptide_from, peptide_to, phase ) = \
              map(int, ( sbjct_genome_from, sbjct_genome_to,
                         peptide_from, peptide_to, phase ) )

            ## sanity check
            if sbjct_genome_to <= sbjct_genome_from:
                ## Ensembl had small exons of wrong orientation.
                ## Those have been removed in later releases.
                ## Simply ignore them here.
                continue
            
            if sbjct_strand in ("-1", "0", "-"):
                sbjct_strand = "-"
            elif sbjct_strand in ( "1", "+", "+1") :
                sbjct_strand = "+"                

            if sbjct_token in self.contigSizes:
                sbjct_length = self.contigSizes[sbjct_token]
                if sbjct_strand == "-":
                    sbjct_genome_from, sbjct_genome_to = sbjct_length - sbjct_genome_to, sbjct_length - sbjct_genome_from
                    
            if last_query != query:
                
                if entry:
                    self.mNumInput += 1
                    try:
                        entry.mMapPeptide2Genome = Genomics.Exons2Alignment( exons )
                    except ValueError, msg:
                        self.mErrors.append( (last_query, msg) )
                    else:
                        self.mNumOutput += 1
                        entry.mAlignmentString = Genomics.Alignment2String( entry.mMapPeptide2Genome )
                        matches.append( entry )
                    exons = []

                    
                id += 1
                entry = PredictionParserEntry()
                
                entry.mRank = 0 
                entry.mQueryToken = query
                entry.mPredictionId = query

                entry.mSbjctToken = sbjct_token
                entry.mSbjctStrand = sbjct_strand
                entry.mSbjctGenomeFrom = sbjct_genome_from
                entry.mSbjctGenomeTo = sbjct_genome_to

                entry.mPercentIdentity = 100
                entry.mPercentSimilarity = 100                
                
                entry.mQueryCoverage = 100                

                last_query = query
                
            else:
                entry.mSbjctGenomeFrom = min(entry.mSbjctGenomeFrom, sbjct_genome_from)
                entry.mSbjctGenomeTo = max(entry.mSbjctGenomeTo, sbjct_genome_to)
                
            exons.append( (sbjct_genome_from, sbjct_genome_to) )
                    
        if entry:
            self.mNumInput += 1
            try:
                entry.mMapPeptide2Genome = Genomics.Exons2Alignment( exons )
            except ValueError, msg:
                self.mErrors.append( (last_query, msg) )
            else:
                entry.score = entry.mSbjctTo - entry.mSbjctFrom                
                entry.mAlignmentString = Genomics.Alignment2String( entry.mMapPeptide2Genome )
                matches.append( entry )
                self.mNumOutput += 1
                
        return matches

class PredictionParserGFF(PredictionParser):
    """Parse GFF files.
    """
    
    def __init__(self, contig_sizes = {}):
        PredictionParser.__init__( self )
        self.contigSizes = contig_sizes
        
    def ExtractData( self,
                     lines, peptide_sequence,
                     genomic_sequence ):
        """extract data from psl, tab-formatted output.
        """
        raise "this module is incomplete"
    
        matches = Predictions()

        for line in lines:
            if line[0] == "#": continue
        
            gff_entry = GTF.Entry()
            gff_entry.read( line )

            entry = PredictionParserEntry()
            entry.mSbjctToken = gff_entry.name
            entry.mSbjctGenomeFrom = gff_entry.start
            entry.mSbjctGenomeTo = gff_entry.end
            
            (query, sbjct_token, sbjct_strand,
             phase, exon_id, peptide_from, peptide_to,
             sbjct_genome_from, sbjct_genome_to ) = data

            ( sbjct_genome_from, sbjct_genome_to,
              peptide_from, peptide_to, phase ) = \
            map(int, ( sbjct_genome_from, sbjct_genome_to,
                       peptide_from, peptide_to, phase ) )
            
            if sbjct_strand in ("-1", "0"):
                sbjct_strand = "-"
            elif sbjct_strand == "1":
                sbjct_strand = "+"                

            if sbjct_token in self.contigSizes:
                sbjct_length = self.contigSizes[sbjct_token]
                if sbjct_strand == "-":
                    sbjct_genome_from, sbjct_genome_to = sbjct_length - sbjct_genome_to, sbjct_length - sbjct_genome_from

            if last_query != query:

                if entry:
                    entry.mMapPeptide2Genome = Genomics.Exons2Alignment( exons )
                    entry.mAlignmentString = Genomics.Alignment2String( entry.mMapPeptide2Genome )

                    matches.append( entry )
                    exons = []
                    
                id += 1
                entry = PredictionParserEntry()
                
                entry.mRank = 0 
                entry.mQueryToken = query
                entry.mPredictionId = id

                entry.mSbjctToken = sbjct_token
                entry.mSbjctStrand = sbjct_strand
                entry.mSbjctGenomeFrom = sbjct_genome_from
                entry.mSbjctGenomeTo = sbjct_genome_to

                entry.mPercentIdentity = 100
                entry.mPercentSimilarity = 100                
                
                entry.mQueryCoverage = 100                
                
                
                last_query = query
                
            else:
                entry.mSbjctGenomeFrom = min(entry.mSbjctGenomeFrom, sbjct_genome_from)
                entry.mSbjctGenomeTo = max(entry.mSbjctGenomeTo, sbjct_genome_to)

            exons.append( (sbjct_genome_from, sbjct_genome_to) )
                    
        if entry:
            entry.mMapPeptide2Genome = Genomics.Exons2Alignment( exons )
            entry.mAlignmentString = Genomics.Alignment2String( entry.mMapPeptide2Genome )

            matches.append( entry )
            
        return matches


if __name__ == "__main__":

    p = PredictionParserGenewise()
    p.Parse( sys.stdin.readlines() )
    print str(p)
                                                      



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
WrapperMEDIPS.py - wrap MEDIPS methylation analysis
===================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

perform Medip-Seq data analysis using the MEDIPS R package.

Note that the UCSC genome file has to have been downloaded 
previously in Bioconductor::

   source("http://bioconductor.org/biocLite.R")
   biocLite("BSgenome.Hsapiens.UCSC.hg19")


Usage
-----

Documentation
-------------

Code
----

'''

import os, sys, re, optparse, tempfile, shutil, subprocess, tempfile

import Experiment as E
import IOTools
import IndexedFasta

## for zinba
from rpy2.robjects import r as R
import rpy2.robjects as ro
import rpy2.robjects.vectors as rovectors
from rpy2.rinterface import RRuntimeError


def bamToMEDIPS( infile, outfile ):
    '''convert bam to medips format

    contig, start, end, strand

    Start is 1-based.
    '''

    statement = '''bamToBed -i %(infile)s | awk '{printf("%%s\\t%%i\\t%%i\\t%%s\\n", $1,$2+1,$3,$6)}' > %(outfile)s''' % locals()

    E.debug( "executing statement '%s'" % statement )

    E.run( statement )

    return outfile

def bedToMEDIPS( infile, outfile ):
    '''convert bam to medips format

    contig, start, end, strand

    Start is 1-based.
    '''

    if infile.endswith( ".gz" ): cat = "zcat"
    else: cat = "cat"

    statement = '''%(cat)s %(infile)s | awk '{printf("%%s\\t%%i\\t%%i\\t%%s\\n", $1,$2+1,$3,$6)}' > %(outfile)s''' % locals()

    E.run( statement )

    return outfile

def compress( infile ):
    '''gzip infile'''

    statement = "gzip -f %(infile)s" % locals() 

    E.debug( "executing statement '%s'" % statement )

    return E.run( statement )

def bigwig( infile, contig_sizes ):
    '''convert infile to bigwig file'''

    if infile.endswith( ".wig"):
        outfile = infile[:-4] + ".bigwig"
    else:
        outfile = infile + ".bigwig"
        
    tmp, filename_sizes = tempfile.mkstemp() 

    os.write( tmp, "\n".join( [ "\t".join(map(str,x)) for x in contig_sizes.iteritems() ] ) )
    os.close( tmp )

    statement = "wigToBigWig -clip %(infile)s %(filename_sizes)s %(outfile)s " % locals() 

    E.debug( "executing statement '%s'" % statement )

    if E.run( statement ):
        os.unlink( infile )

    os.unlink( filename_sizes )

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-f", "--input-format", dest="input_format", type="choice",
                      choices = ("bed", "bam"),
                      help="input file format [default=%default]."  )
    
    parser.add_option("-u", "--ucsc-genome", dest="ucsc_genome", type="string",
                      help="UCSC genome identifier [default=%default]."  )

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )

    parser.add_option("-e", "--extension", dest="extension", type="int",
                      help="extension size [default=%default]."  )

    parser.add_option("-b", "--bin-size", dest="bin_size", type="int",
                      help="bin size of genome vector [default=%default]."  )

    parser.add_option("-l", "--fragment-length", dest="fragment_length", type="int",
                      help="bin size of genome vector [default=%default]."  )

    parser.add_option("-s", "--saturation-iterations", dest="saturation_iterations", type="int",
                      help = "iterations for saturation analysis [default=%default]."  )
    
    parser.add_option( "-t", "--toolset", dest="toolset", type="choice", action="append",
                       choices = ("saturation", "coverage", "rms", "rpm", "all"),
                       help = "actions to perform [default=%default]." )
    
    parser.add_option( "-w", "--bigwig", dest="bigwig", action = "store_true",
                       help = "store wig files as bigwig files - requires a genome file [default=%default]" )

    parser.set_defaults(
        input_format = "bam",
        ucsc_genome = "hg19",
        genome_file = None,
        extension = 400,
        bin_size = 50,
        saturation_iterations = 10,
        fragment_length = 700,
        toolset = [],
        bigwig = False,
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    if len(args) != 1:
        raise ValueError("please specify a filename with sample data")

    if options.bigwig and not options.genome_file:
        raise ValueError("please provide a genome file when outputting bigwig")

    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
        contig_sizes = fasta.getContigSizes()
        
    filename_sample = args[0]

    if len(options.toolset) == 0: options.toolset = ["all"]

    do_all = "all" in options.toolset
    
    # load MEDIPS
    R.library( 'MEDIPS' )
    genome_file = 'BSgenome.Hsapiens.UCSC.%s' % options.ucsc_genome 
    R.library( genome_file )
    
    tmpdir = tempfile.mkdtemp( )

    E.debug( "temporary files are in %s" % tmpdir )

    bin_size = options.bin_size
    extension = options.extension
    fragment_length = options.fragment_length
    saturation_iterations = options.saturation_iterations

    if options.input_format == "bam":
        E.info( "converting bam files" )
        filename_sample = bamToMEDIPS( filename_sample, os.path.join( tmpdir, "sample.medips" ) )
    elif options.input_format == "bed":
        E.info( "converting bed files" )
        filename_sample = bedToMEDIPS( filename_sample, os.path.join( tmpdir, "sample.medips" ) )

    E.info( "loading data" )
    R('''CONTROL.SET = MEDIPS.readAlignedSequences(
                       BSgenome = "%(genome_file)s", 
                       file = "%(filename_sample)s" ) ''' % locals() )
    slotnames = ( ( "extend", "extend", "%i"),
                  ( "distFunction", "distance_function", "%s"),
                  ( "slope", "slope", "%f"),
                  ( "fragmentLength", "fragment_length", "%i" ),
                  ( "bin_size", "bin_size", "%i"),
                  ( "seq_pattern", "pattern", "%s" ),
                  ( "number_regions", "nregions", "%i"),
                  ( "number_pattern", "npatterns", "%i" ),
                  ( "cali_chr", "calibration_contig", "%s"),
                  ( "genome_name", "genome", "%s") )


    E.info( "computing genome vector" )
    R('''CONTROL.SET = MEDIPS.genomeVector(data = CONTROL.SET, 
                       bin_size = %(bin_size)i, 
                       extend=%(extension)i )''' % locals())

    E.info( "computing CpG positions" )
    R('''CONTROL.SET = MEDIPS.getPositions(data = CONTROL.SET, pattern = "CG")''' )

    E.info( "compute coupling vector" )
    R('''CONTROL.SET = MEDIPS.couplingVector(data = CONTROL.SET, 
                       fragmentLength = %(fragment_length)i, 
                       func = "count")''' % locals() )
    
    E.info( "compute calibration curve" )
    R('''CONTROL.SET = MEDIPS.calibrationCurve(data = CONTROL.SET)''')

    E.info( "normalizing" )
    R('''CONTROL.SET = MEDIPS.normalize(data = CONTROL.SET)''')

    outfile = IOTools.openFile( E.getOutputFile( "summary.tsv.gz" ), "w" )
    outfile.write( "category\tvalue\n" )

    if "saturation" in options.toolset or do_all:
        E.info( "saturation analysis" )
        R('''sr.control = MEDIPS.saturationAnalysis(data = CONTROL.SET, 
                            bin_size = %(bin_size)i, 
                            extend = %(extension)i, 
                            no_iterations = %(saturation_iterations)i, 
                            no_random_iterations = 1)''' % locals() )

        R.png( E.getOutputFile( "saturation.png" ) )
        R('''MEDIPS.plotSaturation(sr.control)''')
        R('''dev.off()''')

        R('''write.csv( sr.control$estimation, file ='%s' )'''% E.getOutputFile( "saturation_estimation.csv" ) )
        outfile.write( "estimated_correlation\t%f\n" % R('''sr.control$maxEstCor''')[1] )
        outfile.write( "true_correlation\t%f\n" % R('''sr.control$maxTruCor''')[1] )

    if "coverage" in options.toolset or do_all:
        E.info( "CpG coverage analysis" )
        R('''cr.control = MEDIPS.coverageAnalysis(data = CONTROL.SET, 
                                extend = %(extension)i, 
                                no_iterations = 10)''' % locals())

        R.png( E.getOutputFile( "cpg_coverage.png" ) )
        R('''MEDIPS.plotCoverage(cr.control)''')
        R('''dev.off()''')

        # three rows
        R('''write.csv( cr.control$coveredPos, file ='%s' )'''% E.getOutputFile( "saturation_coveredpos.csv" ) )
        # coverage threshold
        # number of CpG covered
        # percentage of CpG covered

        R('''write.csv( cr.control$matrix, file ='%s' )'''% E.getOutputFile( "saturation_matrix.csv" ) )

        # R('''er.control = MEDIPS.CpGenrich(data = CONTROL.SET)''')

    if "calibration" in options.toolset or do_all:
        E.info( "plotting calibration" )
        R.png( E.getOutputFile( "calibration.png" ) )
        R('''MEDIPS.plotCalibrationPlot(data = CONTROL.SET, linearFit = T, xrange=250)''')
        R('''dev.off()''')

    
    for slotname, label, pattern in slotnames:
        value = tuple(R('''CONTROL.SET@%s''' % slotname ))
        if len(value) == 0: continue
        outfile.write( "%s\t%s\n" % (label, pattern % tuple(R('''CONTROL.SET@%s''' % slotname ))[0] ) )
        
    outfile.close()
        
    if "rpm" in options.toolset or do_all:
        outputfile = E.getOutputFile( "rpm.wig" )
        R('''MEDIPS.exportWIG(file = '%(outputfile)s', data = CONTROL.SET, raw = T, descr = "rpm")''' % locals())
        if options.bigwig:
            bigwig( outputfile, contig_sizes )
        else:
            compress( outputfile )
    
    if "rms" in options.toolset or do_all:
        outputfile = E.getOutputFile( "rms.wig" )
        R('''MEDIPS.exportWIG(file = '%(outputfile)s', data = CONTROL.SET, raw = F, descr = "rms")''' % locals())
        if options.bigwig:
            bigwig( outputfile, contig_sizes )
        else:
            compress( outputfile )

    shutil.rmtree( tmpdir )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
ProfileLibrary.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import os, sys, string, re, tempfile, subprocess, optparse, time, math

#--------------------------------------------------------
#--------------------------------------------------------
#--------------------------------------------------------
# usage message
#--------------------------------------------------------
USAGE="""python %s [OPTIONS]

work with library of profiles.

Actions include: 
   * create: create a profile library
   * verify: verify a profile library
   * merge:  merge several profile libraries
   * split:  split a profile library into smaller sections

This is version $Id: ProfileLibrary.py 2781 2009-09-10 11:33:14Z andreas $.
""" % sys.argv[0]

#--------------------------------------------------------
#--------------------------------------------------------
#--------------------------------------------------------
# import of user libraries
#--------------------------------------------------------
import Experiment
import Mali
try: import alignlib
except ImportError: pass

class ProfileLibrary:

    
    mSuffixDatabase = ".pdb"
    mSuffixIndex = ".pix"

    def __init__(self, name, mode = "r" ):
        self.mName = name

        self.mFilenameProfiles = self.mName + self.mSuffixDatabase
        self.mFilenameIndex = self.mName + self.mSuffixIndex
        self.mIndex = {}
        self.mWeightor = None
        self.mRegularizor = None

        self.mOutfileDatabase = None
        self.mOutfileIndex = None

        if mode == "r":
            self.__loadIndex()
        elif mode == "w":
            if os.path.exists( self.mFilenameProfiles):
                raise "profile database %s already exists." % self.mFilenameProfiles
            self.mOutfileDatabase = open( self.mFilenameProfiles, "wb" )
            self.mOutfileIndex = open( self.mFilenameIndex, "wb" )
        else:
            raise "unknown mode '%s'" % mode
        
    def __getitem__(self, key):
        return self.getProfile( key )

    def __len__(self):
        return len(self.mIndex )

    def __iter__(self):
        for name in self.mIndex.keys():
            yield (name, self[name])
            
    def keys(self):
        return list(self.iterkeys())

    def iterkeys( self ):
        for name in self.mIndex.iterkeys():
            yield name

    def iteritems( self ):
        for name in self.mIndex.keys():
            yield (name, self[name])

    def __del__(self):

        if self.mOutfileDatabase:
            self.mOutfileDatabase.close()
        if self.mOutfileIndex:
            self.mOutfileIndex.close()

    def setOptions( self, options ):
        """set options - access to command line options."""
        self.mLogLevel = options.loglevel
        self.mStdOut = options.stdout
        self.mStdLog = options.stdlog
        self.mStdErr = options.stderr
        
    def setWeightor( self, weightor ):
        """set the sequence weightor to use for profile creation."""
        if weightor == None or weightor == "none":
            self.mWeightor = alignlib.makeNoWeightor() 
        elif weightor == "Henikoff" :
            self.mWeightor = alignlib.makeWeightorHenikoff()
        elif weightor == "HenikoffKimmen":
            self.mWeightor = alignlib.makeWeightorHenikoffKimmen()

    def __loadIndex( self ):

        if not os.path.exists( self.mFilenameProfiles):
            raise "profile database %s could not be found." % self.mFilenameProfiles

        if not os.path.exists( self.mFilenameIndex):
            raise "index %s could not be found." % self.mFilenameIndex
        
        infile = open( self.mFilenameIndex, "r" )
        self.mIndex = {}

        for line in infile:
            if line[0] == "#": continue
            name, first_pos, last_pos = line[:-1].split("\t")
            self.mIndex[name] = (int(first_pos), int(last_pos) )
            
        self.mInfileDatabase = open( self.mFilenameProfiles, "rb" )

    def appendProfile( self, name, profile ):
        """append a profile to this library."""

        start = self.mOutfileDatabase.tell()
        profile.save( self.mOutfileDatabase )
        self.mOutfileIndex.write( "%s\t%s\t%s\n" % (name, 
                                                    str(start),
                                                    str(self.mOutfileDatabase.tell()) ))

    def getProfile( self, name ):
        """append a profile to this library."""

        if name not in self.mIndex: raise KeyError

        self.mInfileDatabase.seek( self.mIndex[name][0] )
        return alignlib.loadAlignandum( self.mInfileDatabase )

    def create( self, infile ):
        """create profile library from file."""

        self.mOutfileDatabase = open( self.mFilenameProfiles, "wb" )
        outfile_index = open( self.mFilenameIndex, "w" )

        ninput, noutput = 0, 0

        while mali.readFromFile( sys.stdin, format="profile" ):

            ninput += 1

            m = Mali.convertMali2Alignlib( mali )
            p = alignlib.makeProfile( m, weightor = self.mWeightor )
            p.prepare()

            self.appendProfile( mali.getName(), p )
            
            noutput += 1

        return ninput, noutput

    def verify( self, infile ):
        """verify data in database against original data."""

        if not self.mIndex: self.__loadIndex()
        
        ninput, nfound, nnotfound, ndifferent = 0,0,0,0
        while mali.readFromFile( sys.stdin, format="profile" ):

            ninput += 1
            m = Mali.convertMali2Alignlib( mali )
            p1 = alignlib.makeProfile( m )
            p1.prepare()
            
            p2 = self.getProfile( mali.getName() )

            if p1.getLength() != p2.getLength() or \
                    str(p1) != str(p2):
                ndifferent += 1
                continue
            
            nfound += 1

        return ninput, nfound, nnotfound, ndifferent

#--------------------------------------------------------
#--------------------------------------------------------
#--------------------------------------------------------
# main part of script
#--------------------------------------------------------
if __name__ == "__main__":

    #--------------------------------------------------------
    # command line parsing options
    parser = E.OptionParser( version = "%prog version: $Id: ProfileLibrary.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-p", "--prefix", dest="prefix", type="string",
                      help="prefix to use for the profile library." )

    parser.add_option("-a", "--action", dest="action", type="choice",
                      choices=("create", "verify", "merge", "split", "stats" ),
                      help="action to undertake." )

    parser.add_option("-w", "--weightor", dest="weightor", type="choice",
                      choices=("none", "Henikoff", "HenikoffKimmen" ),
                      help="sequence weightor to choose." )

    parser.set_defaults( prefix = "profiles", 
                         action = "create",
                         weightor = None,
                         )

    (options, args) = Experiment.Start( parser )

    #--------------------------------------------------------
    # main part of script
    mali = Mali.Mali()

    if options.action in ("create", "merge" ):
        mode = "w"
    else:
        mode = "r"

    plib = ProfileLibrary( options.prefix, mode )
    plib.setWeightor( options.weightor )

    if options.action == "verify":
        ninput, nfound, nnotfound, ndifferent = plib.verify(sys.stdin)
        if options.loglevel >= 1:
            options.stdlog.write( "# verify: ninput=%i, nfound=%i, nnotfound=%i, ndifferent=%i\n" % (ninput, nfound, nnotfound, ndifferent))

    elif options.action == "create":
        # create a new profile library
        ninput, noutput = plib.create( sys.stdin )
        if options.loglevel >= 1:
            options.stdlog.write( "# ninput=%i, noutput=%i\n" % (ninput, noutput) )

    elif options.action == "merge":
        for library in args:
            if options.loglevel >= 1:
                options.stdlog.write("# adding library %s\n" % library )
                options.stdlog.flush()

            other_lib = ProfileLibrary( library )

            for name, profile in other_lib.iteritems():
                plib.appendProfile( name, profile )

    elif options.action == "stats":
        options.stdout.write("profiles\t%i\n" % len(plib) )
        
    #--------------------------------------------------------
    # general cleaning up
    Experiment.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: chain2stats.py 10 2011-04-21 11:07:33Z steve $
#
#   Copyright (C) 2011 Stephen Sansom
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
chain2stats.py
==============

:Author: Stephen Sansom
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   This script computes the coverage of the target and query genomes represented by a UCSC chain file. 

Usage
-----

Example::

   python chain2stats.py -c chainfile

Type::

   python chain2stats.py --help

for command line help.

Documentation
-------------

For large chain file this script requires a substantial amount of memory for chromosome pair analysis.

Code
----
'''

#import modules
import os
import sys
from optparse import OptionParser
from operator import add
from numpy import *
from operator import itemgetter,attrgetter
import bx.bitset
import bx.bitset_builders
import collections

import CGAT.Experiment as E
from CGAT.IndexedFasta import IndexedFasta

############################ Functions/Generators ############################

def chain_iterator(infile):
    lines = []
    for line in open(infile,'r'):
        if line.startswith("#"): continue
        if line.strip() == "": continue 
        if line.startswith("chain"):
            if lines: yield lines
            lines = []
        lines.append( line )
    yield lines

######################### Functions/Generators end ############################


###############################################################################
################################ Classes ######################################
###############################################################################

##-----------------------------------------------------------------------------

class Chain:

    "class with methods for getting gapped and ungapped regions represented by a chain"

    def __init__(self,data):
        self.lines = data
        self.header = self.lines[0]
        self.atts = self.header.split()
        a = len(self.atts)
        if a != 13: #sanity check, can remove
            raise Exception("Unexpected Header length\n"+self.header)
        self.score = int(self.atts[1])
        self.tname = self.atts[2]
        self.tsize = int(self.atts[3])
        self.tstrand = self.atts[4]
        self.tstart = int(self.atts[5])
        self.tend = int(self.atts[6])
        self.qname = self.atts[7]
        self.qsize = int(self.atts[8])
        self.qstrand = self.atts[9]
        b = self.qstrand
        if (b != "+") and (b != "-"):
            raise Exception("unexpected strand id")
        self.qstart = int(self.atts[10])
        self.qend = int(self.atts[11])
        self.id = int(self.atts[12])
        self.tlength = self.tend - self.tstart
        self.qlength = self.qend - self.qstart
        self.nlines = len(self.lines)
        self.nblocks = self.nlines - 1
        (self.tgr,self.qgr) = self._gapped_regions()
        (self.tugr,self.qugr) = self._ungapped_regions()

    def _gapped_regions(self):
        tr,qr = [],[]
        tr.append([self.tstart,self.tlength])
        qr.append([self.qstart,self.qlength])
        return(tr,qr)

    def _ungapped_regions(self):
        ts = self.tstart
        qs = self.qstart
        tr,qr = [],[]
        for i in range(1,self.nlines):
            dataline = self.lines[i]
            values = dataline.split()
            len_align = int(values[0])
            te = ts + len_align
            qe = qs + len_align
            tr.append([ts,len_align])
            qr.append([qs,len_align])
            if (i < self.nlines-1):
                ts = te + int(values[1])
                qs = qe + int(values [2])
        return(tr,qr)

    def get_regions(self,gapped):
        if gapped == True:
            treg, qreg = self.tgr, self.qgr
        else:
            treg, qreg = self.tugr, self.qugr            
        return(treg,qreg)

##-----------------------------------------------------------------------------

class Bitsets_Container:

    "General class for holding bitsets in a {key1: {key2: bitset }} structure"

    def __init__(self):
        self.bitsets = collections.defaultdict( dict )

    def addRanges(self,key1,key2,size,ranges):
        if key2 not in self.bitsets[key1]:
            if size == 199501827: # this randomly segfaults a binnedbitset
                self.bitsets[key1][key2] = bx.bitset.BitSet(size)
            else:
                self.bitsets[key1][key2] = bx.bitset.BinnedBitSet(size) 
        for r in ranges: #bx-python does sanity checks.
           self.bitsets[key1][key2].set_range(r[0],r[1]) 

    def bitset_coverage(self,key1,key2):
        b = self.bitsets[key1][key2]
        l = b.size
        c = b.count_range( 0, l )
        return (c , l)

##**---------------------------------------------------------------------------

class ChainCounter():

    '''
    Top, defining class for a chain counter.
    
    Counters must have:
    (1) an add method to add a new chain
    (2) a write_report method

    Counters will generally initialise a container and add metrics to it when passed a chain
    '''

#---------- abstract methods------------

    def add(self,c):
        raise NotImplementError("abstract method - implement in subclass")

    def _get_stats(self,options):
        '''Calculates and stores the stats if necessary - called by all report methods'''
        raise NotImplementError("abstract method - implement in subclass")

    def report(self,options):
        '''should call the "write_report" method to do the writing'''
        raise NotImplementError("abstract method - implement in subclass")

    def tabbed_report(self,options):
        '''should call the "write_tabbed" method to do the writing'''
        raise NotImplementError("abstract method - implement in subclass")


#----------- methods --------------------

    def _get_basic_stats(self,numlist,string="{:.2f}"):
        s = [string.format(mean(numlist)),string.format(median(numlist)),string.format(max(numlist)),string.format(min(numlist))]
        return(s)

    def _wrap_basic_stats(self,s,tabbed=False):
        if tabbed==False:
            r = ''.join(["Mean:",s[0],", median:",s[1],", max:",s[2],", min:",s[3],"\n"])
        else: 
            r = '\t'.join(s)
        return(r)

    def _write_report(self,options,report):
        options.stdout.write(''.join(['\n','\n'.join(report),'\n']))
        
    def _write_tabbed(self,name,lines,E):
        outfile = E.openOutputFile(name)
        outfile.write('\n'.join(lines))
        outfile.write('\n')
        outfile.close

    def _wrap_header(self,text=""):
        '''decorates header for report'''
        if text == "": text = self.header
        underline = ''.join(["-" for i in text])
        return([text,underline])

##----------------------------------------------------------------------------

class CounterPerChromosome(ChainCounter):

    header = "Report for chains per chromosome"
    
    def __init__(self,gapped):
        self.container = Bitsets_Container()
        self.gapped = gapped
        self.cov = 0
        if self.gapped == True: self.name="gapped"
        else: self.name = "ungapped"

    def _pc_coverage(self,coverage,total_length):
        pc_cov = float(coverage)/float(total_length) * 100
        return(pc_cov)

    def _coverage(self,key1):
        total, total_len = 0,0
        d = self.container.bitsets[key1]
        for key2 in sorted(d):
            (s,l) = self.container.bitset_coverage(key1,key2)
            total += s
            total_len += l
        gen_coverage = self._pc_coverage(total,total_len)
        return gen_coverage

    def add(self,c):
        treg, qreg = c.get_regions(self.gapped)
        self.container.addRanges("target",c.tname,c.tsize,treg)
        self.container.addRanges("query",c.qname,c.qsize,qreg)

    def _get_stats(self):
        if self.cov == 0: 
            self.cov = {}
            for i in ("target","query"):
                self.cov[i] = self._coverage(i)

    def report(self,options):
        self._get_stats()
        report = self._wrap_header("".join([self.header," ",self.name," statistics"]))
        report.append("Target genome {:.2f}, query genome: {:.2f}\n".format(self.cov["target"],self.cov["query"]))
        self._write_report(options,report)

    def tabbed_report(self,options,E):
        self._get_stats()
        lines = ["target_genome\tquery_genome"]
        lines.append('{:.2f}\t{:.2f}'.format(self.cov["target"],self.cov["query"]))
        self._write_tabbed(''.join(["genome_coverage_",self.name]),lines,E)        

##-----------------------------------------------------------------------------

class CounterPerChromosomePair(CounterPerChromosome):
    
    header = "Report for chains per chromosome pair"

    def _make_table(self,options):
        table = []
        table.append("t_chrom\tt_cov\tq_chrom\tq_cov\tbases aligned")
        tchrom = self.cov[0][0]
        j = 0
        for i in self.cov:
            if i[0]!=tchrom:
                tchrom = i[0]
                j=0       
            if j < options.nperchrom:
                table.append("{}\t{:.2f}\t{}\t{:.2f}\t{:,}".format(i[0],i[1],i[2],i[3],i[4]))
                j += 1
            else:
                continue
        table.append("")
        return(table)

    def add(self,c):
        treg, qreg = c.get_regions(self.gapped)
        if c.qstrand == "+": strand = "plus" 
        else: strand = "minus"
        pair_name = (c.tname, c.qname)
        self.container.addRanges(pair_name,"target",c.tsize,treg)
        self.container.addRanges(pair_name,"query",c.qsize,qreg)
        self.container.addRanges(pair_name,strand,c.qsize,qreg)

    def _get_stats(self,options):
        if self.cov == 0:
            result = []
            for key1 in self.container.bitsets:
                (tm,tl) = self.container.bitset_coverage(key1,'target')
                tc = self._pc_coverage(tm,tl)
                (qm,ql) = self.container.bitset_coverage(key1,'query')
                qc = self._pc_coverage(qm,ql)
                result.append([key1[0],tc,key1[1],qc,tm])
            a = sorted(result,key=itemgetter(4),reverse=True)
            self.cov = sorted(a,key=itemgetter(0))
            self.cov_table = self._make_table(options)

    def report(self,options):
        self._get_stats(options)
        report = (self._wrap_header("".join([self.header," ",self.name," statistics"])))
        report = report + self.cov_table
        self._write_report(options,report)

    def tabbed_report(self,options,E):
        self._get_stats(options)
        self._write_tabbed(''.join(["chrom_pair_cov_",self.name]),self.cov_table,E)        

##-----------------------------------------------------------------------------

class CounterOfGappedChainLengths( ChainCounter ):
    
    header = "Report for chain lengths"

    def __init__(self,gapped):
        self.nchains = 0
        self.tcls,self.qcls = [],[]
        self.stats = {}
        self.gapped = gapped
        if self.gapped == True: self.name="gapped"
        else: self.name = "ungapped"

    def _chain_stats_report(self,name):
        report = [''.join(["Stats for ",name," chain lengths:"])]
        report.append(self._wrap_basic_stats(self.stats[name]))
        return(report)      

    def add(self,c):
        treg, qreg = c.get_regions(self.gapped)
        n = len(treg)
        if n != len(qreg):
            raise Exception("Different numbers of chains for query and target!?!")
        for i in range(0,n):
            self.tcls.append(treg[i][1])
            self.qcls.append(qreg[i][1])
            self.nchains += 1

    def _get_stats(self):
        if ("target" in self.stats)==0:
            self.stats["target"] = self._get_basic_stats(self.tcls,string="{:,.0f}")
        if ("query" in self.stats)==0:
            self.stats["query"] = self._get_basic_stats(self.qcls, string="{:,.0f}")

    def report(self,options):
        self._get_stats()
        report = (self._wrap_header(" ".join([self.header,self.name])))
        report.append("number of chains = {:}\n".format(self.nchains))
        for i in ("target","query"): 
            report = report + self._chain_stats_report(i)
        self._write_report(options,report)

    def tabbed_report(self,options,E):
        self._get_stats()
        for i in ("target","query"):
            lines = ["mean\tmedian\tmax\tmin"]
            lines.append(self._wrap_basic_stats(self.stats[i],tabbed=True))
            self._write_tabbed('_'.join([i,"lengths",self.name]),lines,E)        

##-----------------------------------------------------------------------------

class CounterPercentIdentify( ChainCounter ):

    header = "Report on Percent Indentities"

    def __init__(self,tname,qname):
        self.tfasta = IndexedFasta(tname)
        self.qfasta = IndexedFasta(qname)
        self.pids = []
        self.stats = 0

    def _get_pid(self,x,y):
        z = zip(x,y)
        pid = (float(len([x for x,y in z if x==y])) / float(len(z)) * 100)
        return(pid)

    def add(self,c):
        nreg = len(c.tugr)
        for i in range(0,nreg):                                                            
            tseq = self.tfasta.getSequence(c.tname,"+",c.tugr[i][0],(sum(c.tugr[i])))
            qseq = self.qfasta.getSequence(c.qname,c.qstrand,c.qugr[i][0],(sum(c.qugr[i])))
            pid = self._get_pid(tseq.lower(),qseq.lower())
        self.pids.append( pid )
        
    def _get_stats(self):
        if self.stats == 0:
            self.stats = self._get_basic_stats(self.pids,string="{:.2f}")            

    def report(self,options):
        self._get_stats()
        report = self._wrap_header()
        report.append(self._wrap_basic_stats(self.stats))
        self._write_report(options,report)

    def tabbed_report(self,options,E):
        self._get_stats()
        lines = ["mean\tmedian\tmax\tmin"]
        lines.append(self._wrap_basic_stats(self.stats,tabbed=True))
        self._write_tabbed("pids",lines,E)        
       
##-----------------------------------------------------------------------------

class CounterOfErrors( ChainCounter ):

    '''class for reporting invalid contig sizes in chains'''

    header = "Contig size validation report"

    def  __init__(self,options):
        self.tdb = IndexedFasta(options.dbpath+options.targetgenome)
        self.qdb = IndexedFasta(options.dbpath+options.querygenome)
        self.tcontigs = self.tdb.getContigSizes()
        self.qcontigs = self.qdb.getContigSizes()
        self.badchains = []

    def add(self,c):
        db_tsize = self.tcontigs[c.tname]
        db_qsize = self.qcontigs[c.qname]
        if c.tsize != db_tsize:
            self.badchains.append('\t'.join([str(x) for x in c.atts] + [" #bad target contigsize"]))
        if c.qsize != db_qsize:
            self.badchains.append('\t'.join([str(x) for x in c.atts] + [" #bad query contigsize"]))
    
    def report(self,options):
        report = self._wrap_header()
        if len(self.badchains) == 0:
            report.append("All chains passed validation")
        else:
            report = report + self.badchains
        self._write_report(options,report)

    def tabbed_report(self,options,E):
        if len(self.badchains) > 0:
            lines = self.badchains
        else:
            lines = ["#no bad chains found"]
        self._write_tabbed("bad_contig_sizes",lines,E)        

##-----------------------------------------------------------------------------

###############################################################################
############################# Classes end #####################################
###############################################################################


def main(argv = None ):
    if not argv: argv = sys.argv
    
    #get the options
    parser = OptionParser(version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                          usage = globals()["__doc__"] )

    parser.add_option("-c", "--chainfile", dest="chainfile", type="string",
                      help="the chain file to analyse", metavar="FILE")
    parser.add_option("-p", "--perchrom", dest="perchrom", action="store_true",
                      help="Set to 1 or \"True\" to perform per chromosome pair analysis",default=False)
    parser.add_option("-n", "--nperchrom", dest="nperchrom", type="int",
                      help="Number of aligments to report on per chromosome pair",default=2)
    parser.add_option("-i","--identity", dest="identity", action="store_true",
                      help="Generate stats on the sequence identity of the gapped chains. Requires FastIndex.py", default=False)
    parser.add_option("-d", "--dbpath", dest="dbpath", type="string",
                      help="The path to the indexed fasta files",default="/ifs/mirror/genomes/plain/")
    parser.add_option("-t", "--targetgenome", dest="targetgenome", type="string",
                      help="The target genome, eg. Mm19",default=False)
    parser.add_option("-q", "--querygenome", dest="querygenome", type="string",
                      help="The query genome eg. Hg17",default=False)
    parser.add_option("-e", "--errors", dest="errors", action="store_true",
                      help="Check chains for erroneous contig sizes using the given db",default=False)
    parser.add_option("-r", "--report", dest="report", action="store_true",
                      help="Write out tab-delimited reports for each analysis",default=False)



    (options, args) = E.Start(parser, argv = argv, add_output_options = True )
    
    #make a list of counting objects
    counters = []

    counters.append(CounterPerChromosome( gapped = True ) )
    counters.append(CounterPerChromosome( gapped = False ) )

    if options.perchrom:
        counters.append( CounterPerChromosomePair( gapped = True ) )
        counters.append( CounterPerChromosomePair( gapped = False ) )

    counters.append(CounterOfGappedChainLengths( gapped = True ))
    counters.append(CounterOfGappedChainLengths( gapped = False ))

    if options.identity == True:
        if options.targetgenome == 0 or options.querygenome == 0:
            raise Exception("Target and query database must be specified with the \"-e\" flag") 
        t_db_path = os.path.join(options.dbpath,options.targetgenome)
        q_db_path = os.path.join(options.dbpath,options.querygenome)
        counters.append(CounterPercentIdentify( t_db_path, q_db_path ))

    if options.errors == True:
        if options.targetgenome == 0 or options.querygenome == 0:
            raise Exception("Target and query database must be specified with the \"-e\" flag") 
        counters.append(CounterOfErrors(options))

    #iterate over the chains and counters
    for chain in chain_iterator(options.chainfile):
        c = Chain(chain)
        for counter in counters:
            counter.add(c)
    
    #write a report to stdout and individual reports to tab delimited files
    options.stdout.write("\n\n********** chain2stats report starts **********\n")

    for counter in counters:
        counter.report(options)
        if options.report == True:
            counter.tabbed_report(options,E)

    options.stdout.write("\n********** chain2stats report ends **********\n\n")

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
links2fasta.py - convert links into alignments
==============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python links2fasta.py --help

Type::

   python links2fasta.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import tempfile
import time
import optparse

import CGAT.Experiment as E
import CGAT.BlastAlignments as BlastAlignments
import alignlib
import CGAT.Genomics as Genomics
import CGAT.Exons as Exons
import CGAT.FastaIterator as FastaIterator

class Map:
    def __init__(self):
        pass

    def read( self, line ):
        try:
            ( self.mToken,
              self.mOldFrom, self.mOldTo, self.mOldAli,
              self.mNewFrom, self.mNewTo, self.mNewAli,
              self.mOldLength, self.mNewLength) = line[:-1].split("\t")
        except ValueError:
            raise ValueError("parsing error in line\n%s" % line)

        (self.mOldFrom, self.mOldTo, self.mNewFrom, self.mNewTo) = \
                        map(int, (self.mOldFrom, self.mOldTo, self.mNewFrom, self.mNewTo))
        self.mMapOld2New = None
        
    def expand( self ):
        if not self.mMapOld2New:
            self.mMapOld2New = alignlib.makeAlignmentVector()
        
            alignlib.AlignmentFormatEmissions( 
                self.mOldFrom, self.mOldAli,
                self.mNewFrom, self.mNewAli).copy( self.mMapOld2New )
            
    def clear( self ):
        if self.mMapOld2New:
            self.mMapOld2New.clear()
        self.mMapOld2New = None

    def __str__(self):
        return string.join(map(str, (self.mToken,
                                     self.mOldFrom, self.mOldTo, self.mOldAli,
                                     self.mNewFrom, self.mNewTo, self.mNewAli,
                                     self.mOldLength, self.mNewLength)), "\t")
            
def ScaleAlignment( alignment, factor ):
    """scale alignment string."""

    data = re.split("[+-]", alignment[1:])
    
    data = map( lambda x: int(x) * factor, data )
    signs = [ "+", "-" ] * (1 + len(data) / 2)
    
    if alignment[0] == "+":
        del signs[-1]
    else:
        del signs[0]

    s = map( lambda x,y: "%s%i" % (x,y), signs, data)
    return string.join(s, "")

##-------------------------------------------------------------------------
def Write( map_row2col, row_seq, col_seq, link,
           no_gaps = False, no_identical = False,
           min_length = 0,
           suffix1="", suffix2="",
           outfile = None,
           pair_filter = None,
           format = "fasta" ):
    """write alignment based on map_row2col."""
        
    status = None

    filter_status = "new"
    
    if map_row2col.getLength() == 0:
        status = "empty"

    if not status:

        f = alignlib.AlignmentFormatExplicit( map_row2col, row_seq, col_seq )
        
        row_from = map_row2col.getRowFrom()
        row_to = map_row2col.getRowTo()
        col_from = map_row2col.getColFrom()
        col_to = map_row2col.getColTo()
        row_ali, col_ali = f.mRowAlignment, f.mColAlignment

    if not status:
        if no_gaps:
            # remove gaps from fasta
            r = []
            c = []
            for x in range(len(row_ali)):
                if row_ali[x] != "-" and col_ali[x] != "-":
                    r.append( row_ali[x] )
                    c.append( col_ali[x] )
            row_ali = string.join(r, "")
            col_ali = string.join(c, "")            

    if not status and len(row_ali) < min_length:
        status = "length"

    if not status and no_identical:
            if row_ali == col_ali:
                status = "identical"

    if not status:

        if pair_filter:
            id = "%s-%s" % (link.mQueryToken, link.mSbjctToken)
            if id in pair_filter:
                h = Genomics.GetHID( row_ali + ";" + col_ali ) 
                if h in pair_filter[id]:
                    filter_status = "old"
        
        translation1 = Genomics.TranslateDNA2Protein( row_ali )
        translation2 = Genomics.TranslateDNA2Protein( col_ali )        

        if "X" in translation1 or "x" in translation2:
            status = "stops"
        else:
            status = "success"

        if filter_status == "new":
            if format == "fasta":
                print ">%s%s %s %s\n%s\n>%s%s %s %s\n%s" % (link.mQueryToken, suffix1, row_from, row_to, row_ali, 
                                                            link.mSbjctToken, suffix2, col_from, col_to, col_ali )
            elif format == "dummy":
                pass
            else:
                raise ValueError("unknown format")

    if outfile:
        outfile.write( "%s%s\t%s%s\t%s\t%i\t%s\n" % (link.mQueryToken, suffix1, link.mSbjctToken, suffix2,
                                                     status, map_row2col.getLength(), filter_status ) )

    return status

def GetAdjustedBoundaries( id, exons ):
    """return codon adjusted exon boundaries."""

    f, t = exons[id].mPeptideFrom, exons[id].mPeptideTo

    f += exons[id].frame
    
    if id < len(exons) -1:
        next_frame = exons[id+1].frame
    else:
        next_frame = 0
        
    if next_frame:
        t -= 3 - next_frame

    return f, t

if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: links2fasta.py 2446 2009-01-27 16:32:35Z andreas $", usage = globals()["__doc__"] )

    parser.add_option( "-s", "--sequences", dest="filename_sequences", type="string",
                       help="peptide sequence [Default=%default]" )

    parser.add_option( "-f", "--format", dest="format", type="string",
                       help="output format [Default=%default]" )

    parser.add_option( "-e", "--expand",  dest="expand", action="store_true",
                       help="expand positions from peptide to nucleotide alignment [Default=%default]")

    parser.add_option( "-m", "--map",  dest="filename_map", type="string",
                       help="map alignments [Default=%default]")
    
    parser.add_option( "-c", "--codons",  dest="require_codons", action="store_true",
                       help="require codons [Default=%default]")

    parser.add_option( "--one-based-coordinates",  dest="one_based_coordinates", action="store_true",
                       help="expect one-based coordinates. The default are zero based coordinates [Default=%default].")

    parser.add_option( "--no-identical",  dest="no_identical", action="store_true",
                       help="do not output identical pairs [Default=%default]" )

    parser.add_option( "-g", "--no-gaps",  dest="no_gaps", action="store_true",
                       help="remove all gaps from aligned sequences [Default=%default]")

    parser.add_option( "-x", "--exons",  dest="filename_exons", type="string",
                       help="filename with exon boundaries [Default=%default]")
    
    parser.add_option( "-o", "--outfile",  dest="filename_outfile", type="string",
                       help="filename to save links [Default=%default]")

    parser.add_option( "--min-length",  dest="min_length", type="int",
                       help="minimum length of alignment [Default=%default]")

    parser.add_option( "--filter",  dest="filename_filter", type="string",
                       help="given a set of previous alignments, only write new pairs [Default=%default].")

    parser.set_defaults(
        filename_sequences = None,
        filename_exons = None,
        filename_map = None,
        filename_outfile = None,
        no_gaps = False,
        format = "fasta",
        expand = False,
        require_codons = False,
        no_identical = False,
        min_length = 0,
        report_step = 100,
        one_based_coordinates = False,
        filename_filter = None)

    (options, args) = E.Start( parser, add_mysql_options = True )

    t0 = time.time()
    if options.filename_sequences:
        sequences = Genomics.ReadPeptideSequences( open(options.filename_sequences, "r") )
    else:
        sequences = {}

    if options.loglevel >= 1:
        options.stdlog.write( "# read %i sequences\n" % len(sequences) )
        sys.stdout.flush()

    if options.filename_exons:
        exons = Exons.ReadExonBoundaries( open(options.filename_exons, "r") )
    else:
        exons = {}

    if options.loglevel >= 1:
        options.stdlog.write( "# read %i exons\n" % len(exons) )
        sys.stdout.flush()

    if options.filename_map:
        map_old2new = {}
        for line in open(options.filename_map, "r"):
            if line[0] == "#": continue
            m = Map()
            m.read( line )
            map_old2new[m.mToken] = m
    else:
        map_old2new = {}

    if options.loglevel >= 1:
        options.stdlog.write( "# read %i maps\n" % len(map_old2new) )
        sys.stdout.flush()

    if options.filename_filter:
        if options.loglevel >= 1:        
            options.stdlog.write( "# reading filtering information.\n" )
            sys.stdout.flush()
            
        map_pair2hids = {}

        if os.path.exists( options.filename_filter ):
            
            infile = open(options.filename_filter, "r")

            iterator = FastaIterator.FastaIterator( infile )

            while 1:
                cur_record = iterator.next()
                if cur_record is None: break

                record1 = cur_record

                cur_record = iterator.next()
                if cur_record is None: break

                record2 = cur_record

                identifier1 = re.match("(\S+)", record1.title).groups()[0]
                identifier2 = re.match("(\S+)", record2.title).groups()[0]

                id = "%s-%s" % (identifier1, identifier2)
                s = Genomics.GetHID(record1.sequence + ";" + record2.sequence)

                if id not in map_pair2hids: map_pair2hids[id] = []

                map_pair2hids[id].append( s )

            infile.close()
            
        if options.loglevel >= 1:        
            options.stdlog.write( "# read filtering information for %i pairs.\n" % len(map_pair2hids) )
            sys.stdout.flush()
    else:
        map_pair2hids = None
        
    if options.loglevel >= 1:
        options.stdlog.write( "# finished input in %i seconds.\n" % (time.time() - t0))

    if options.filename_outfile:
        outfile = open(options.filename_outfile, "w")
    else:
        outfile = None
        
    map_row2col = alignlib.makeAlignmentVector()
    tmp1_map_row2col = alignlib.makeAlignmentVector()
    counts = {}

    iterations = 0

    t1 = time.time()
    ninput, nskipped, noutput = 0, 0, 0

    for link in BlastAlignments.iterator_links( sys.stdin ):

        iterations += 1
        ninput += 1

        if options.loglevel >= 1:
            if (iterations % options.report_step == 0):
                options.stdlog.write( "# iterations: %i in %i seconds.\n" % (iterations, time.time() - t1) )
                sys.stdout.flush()
                
        if link.mQueryToken not in sequences or \
           link.mSbjctToken not in sequences:
            nskipped += 1
            continue

        if options.loglevel >= 3:
            options.stdlog.write( "# read link %s\n" %  str(link) )
            
        row_seq = alignlib.makeSequence( sequences[link.mQueryToken] )
        col_seq = alignlib.makeSequence( sequences[link.mSbjctToken] )

        if options.one_based_coordinates:
            link.mQueryFrom -= 1
            link.mSbjctFrom -= 1

        if options.expand:
            link.mQueryFrom = link.mQueryFrom * 3 
            link.mSbjctFrom = link.mSbjctFrom * 3
            link.mQueryAli = ScaleAlignment( link.mQueryAli, 3 )
            link.mSbjctAli = ScaleAlignment( link.mSbjctAli, 3 )            
            
        map_row2col.clear()

        alignlib.AlignmentFormatEmissions(
            link.mQueryFrom, link.mQueryAli,
            link.mSbjctFrom, link.mSbjctAli ).copy(  map_row2col )
        
        if link.mQueryToken in map_old2new:
            tmp1_map_row2col.clear()
            map_old2new[link.mQueryToken].expand()
            if options.loglevel >= 3:
                options.stdlog.write( "# combining in row with %s\n" %\
                                      str(alignlib.AlignmentFormatEmissions(map_old2new[link.mQueryToken].mMapOld2New ) ))

            alignlib.combineAlignment( tmp1_map_row2col,
                                      map_old2new[link.mQueryToken].mMapOld2New,
                                      map_row2col,
                                      alignlib.RR )
            map_old2new[link.mQueryToken].clear()
            alignlib.copyAlignment( map_row2col, tmp1_map_row2col )

        if link.mSbjctToken in map_old2new:
            tmp1_map_row2col.clear()
            map_old2new[link.mSbjctToken].expand()            
            if options.loglevel >= 3:
                options.stdlog.write( "# combining in col with %s\n" %\
                                      str(alignlib.AlignmentFormatEmissions(map_old2new[link.mSbjctToken].mMapOld2New ) ))

            alignlib.combineAlignment( tmp1_map_row2col,
                                       map_row2col,
                                       map_old2new[link.mSbjctToken].mMapOld2New,
                                       alignlib.CR )
            map_old2new[link.mSbjctToken].clear()
            alignlib.copyAlignment( map_row2col, tmp1_map_row2col )

        dr = row_seq.getLength() - map_row2col.getRowTo() 
        dc = col_seq.getLength() - map_row2col.getColTo() 
        if dr < 0 or dc < 0:
            raise ValueError("out of bounds alignment: %s-%s: alignment out of bounds. row=%i col=%i ali=%s" %\
                                          (link.mQueryToken,
                                           link.mSbjctToken,
                                           row_seq.getLength(),
                                           col_seq.getLength(),
                                           str(alignlib.AlignmentFormatEmissions(map_row2col))))
            

        if options.loglevel >= 2:
            options.stdlog.write( str( alignlib.AlignmentFormatExplicit( map_row2col, 
                                                                         row_seq, 
                                                                         col_seq )) + "\n" )
        ## check for incomplete codons
        if options.require_codons:

            naligned = map_row2col.getNumAligned()
            
            # turned off, while fixing alignlib
            if naligned % 3 != 0:
                options.stdlog.write( "# %s\n" % str(map_row2col) )
                options.stdlog.write( "# %s\n" % str(link) )
                options.stdlog.write( "# %s\n" % str(map_old2new[link.mQueryToken]) )
                options.stdlog.write( "# %s\n" % str(map_old2new[link.mSbjctToken]) )
                options.stdlog.write( "#\n%s\n" % alignlib.AlignmentFormatExplicit( map_row2col, 
                                                                                    row_seq,
                                                                                    col_seq ) )

                raise ValueError("incomplete codons %i in pair %s - %s" % (naligned, link.mQueryToken, link.mSbjctToken))

        ## if so desired, write on a per exon level:
        if exons:
            if link.mQueryToken not in exons:
                raise IndexError("%s not found in exons" % (link.mQueryToken))
            if link.mSbjctToken not in exons:
                raise IndexError("%s not found in exons" % (link.mSbjctToken))
            exons1 = exons[link.mQueryToken]
            exons2 = exons[link.mSbjctToken]

            ## Get overlapping segments
            segments = Exons.MatchExons( map_row2col, exons1, exons2 )
            
            for a,b in segments:
                tmp1_map_row2col.clear()

                # make sure you got codon boundaries. Note that frameshifts
                # in previous exons will cause the codons to start at positions
                # different from mod 3. The problem is that I don't know where
                # the frameshifts occur exactly. The exon boundaries are given
                # with respect to the cds, which include the frame shifts.
                # Unfortunately, phase information seems to be incomplete in the input files.

                from1, to1 = GetAdjustedBoundaries( a, exons1 )
                from2, to2 = GetAdjustedBoundaries( b, exons2 )

                alignlib.copyAlignment( tmp1_map_row2col, map_row2col,
                                       from1+1, to1, from2+1, to2 )
                
                mode = Write( tmp1_map_row2col, row_seq, col_seq, link,
                              no_gaps = options.no_gaps,
                              no_identical = options.no_identical,
                              min_length = options.min_length,
                              suffix1="_%s" % str(a),
                              suffix2="_%s" % str(b),
                              outfile = outfile,
                              pair_filter = map_pair2hid,
                              format = options.format )

                if mode not in counts: counts[mode] = 0
                counts[mode] += 1

        else:
            mode = Write( map_row2col, row_seq, col_seq, link,
                          min_length = options.min_length,                          
                          no_gaps = options.no_gaps,
                          no_identical = options.no_identical,
                          outfile = outfile,
                          pair_filter = map_pair2hids,
                          format = options.format )
            
            if mode not in counts: counts[mode] = 0
            counts[mode] += 1

        noutput += 1
        
    if outfile: outfile.close()
    
    if options.loglevel >= 1:
        options.stdlog.write("# %s\n" % ", ".join( map( lambda x,y: "%s=%i" % (x,y), counts.keys(), counts.values() ) ))
        options.stdlog.write("# ninput=%i, noutput=%i, nskipped=%i\n" % (ninput, noutput, nskipped) )

    E.Stop()
